"""
https://docs.github.com/en/rest/repos/contents?apiVersion=2022-11-28
"""
import base64
import json
from typing import Any, List, Dict

import requests
from fundrive.core import DriveSystem
from funsecret import read_secret


class GithubDrive(DriveSystem):
    def __init__(self, *args, **kwargs):
        super(GithubDrive, self).__init__(*args, **kwargs)
        self.base_url = "https://api.github.com"
        self.repo_str = None
        self.headers = {}

    def login(self, repo_str, access_tokens=None, *args, **kwargs) -> bool:
        if access_tokens:
            read_secret(cate1="fundrive", cate2="drives", cate3="fungithub", cate4="access_tokens", value=access_tokens)
        else:
            access_tokens = read_secret(cate1="fundrive", cate2="drives", cate3="fungithub", cate4="access_tokens")
        self.headers = {"Authorization": f"Token {access_tokens}"}
        self.repo_str = repo_str
        return True

    def delete(self, git_path, *args, **kwargs) -> bool:
        requests.get(f"{self.base_url}/repos/{self.repo_str}/contents/{git_path}", headers=self.headers)
        return True

    def get_dir_info(self, git_path, *args, **kwargs) -> Dict[str, Any]:
        data = self.get_dir_list(git_path)
        if len(data) > 0:
            return data[0]
        return {}

    def get_file_info(self, git_path, *args, **kwargs) -> Dict[str, Any]:
        res = requests.get(f"{self.base_url}/repos/{self.repo_str}/contents/{git_path}", headers=self.headers)
        data = res.json()
        if len(data) == 0 or "name" not in data:
            return {}
        return {
            "name": data["name"],
            "path": data["path"],
            "size": data["size"],
            "sha": data["sha"]
            # "time": datetime.strptime(last_modified, "%a, %d %b %Y %H:%M:%S %Z").strftime("%Y-%m-%d %H:%M:%S"),
        }

    def get_file_list(self, git_path, recursive=False, *args, **kwargs) -> List[Dict[str, Any]]:
        all_files = []
        res = requests.get(f"{self.base_url}/repos/{self.repo_str}/contents/{git_path}", headers=self.headers)
        for data in res.json():
            if data["type"] != "dir":
                all_files.append(
                    {
                        "name": data["name"],
                        "path": data["path"],
                        "size": data["size"],
                        "sha": data["sha"]
                        # "time": datetime.strptime(last_modified, "%a, %d %b %Y %H:%M:%S %Z").strftime("%Y-%m-%d %H:%M:%S"),
                    }
                )
            elif recursive:
                all_files.extend(self.get_file_list(data["path"]))
        return all_files

    def get_dir_list(self, git_path, recursive=False, *args, **kwargs) -> List[Dict[str, Any]]:
        all_files = []
        res = requests.get(f"{self.base_url}/repos/{self.repo_str}/contents/{git_path}", headers=self.headers)
        for data in res.json():
            if data["type"] == "dir":
                all_files.append(
                    {
                        "name": data["name"],
                        "path": data["path"],
                        "size": data["size"],
                        # "time": datetime.strptime(last_modified, "%a, %d %b %Y %H:%M:%S %Z").strftime("%Y-%m-%d %H:%M:%S"),
                    }
                )
                if recursive:
                    all_files.extend(self.get_file_list(data["path"]))
        return all_files

    def upload_file(
        self,
        file_path="./cache",
        content=None,
        git_path=None,
        message="committing files",
        branch="master",
        overwrite=False,
        *args,
        **kwargs,
    ) -> bool:
        uri = f"{self.base_url}/repos/{self.repo_str}/contents/{git_path}"
        if content is None:
            content = open(file_path, "r").read()
        if not isinstance(content, str):
            content = json.dumps(content)

        data = {"message": message, "content": base64.b64encode(content.encode("utf-8")).decode(), "branch": branch}
        exist_info = self.get_file_info(git_path)
        if "sha" in exist_info:
            data['sha']=exist_info['sha']
        
        response = requests.put(uri, headers=self.headers, json=data)
        if response.status_code not in (200,201):
            print(response.status_code,response.text)
        return True



import asyncio
import logging

from fundrive.magnet.dht.node import Node
from notetool.tool.log import set_default_level

set_default_level(logging.DEBUG)
import random
from fundrive.magnet.dht.network import Server
from fundrive.magnet.dht.protocol import KRPCProtocol
from fundrive.magnet.dht.utils import digest

serve = Server()

protocol = KRPCProtocol(serve.node, serve.peer_storage, serve.token_storage, serve.ksize, buckets=serve.buckets, )

BOOTSTRAP_NODES = (
    ("router.bittorrent.com", 6881),
    ("dht.transmissionbt.com", 6881),
    ("router.utorrent.com", 6881),
    ("router.utorrent.com", 6881),
    ("router.bittorrent.com", 6881),
    ("dht.transmissionbt.com", 6881),
    ("router.bitcomet.com", 6881),
    ("dht.aelitis.com", 6881),

)


# torrent = asyncio.run(protocol.call_ping(node))


async def sayhi(protocol: KRPCProtocol, address):
    for boot in BOOTSTRAP_NODES:
        node = Node(digest(random.getrandbits(255)), ip=boot[0], port=boot[1])
        result = await protocol.call_ping(node)
        print(result)
        print(result[1] if result[0] else "No response received.")


logging.basicConfig(level=logging.DEBUG)
loop = asyncio.get_event_loop()
loop.set_debug(True)

# Start local UDP server to be able to handle responses
listen = loop.create_datagram_endpoint(KRPCProtocol, local_addr=('127.0.0.1', 4567))
transport, protocol = loop.run_until_complete(listen)

# Call remote UDP server to say hi
# func = sayhi(protocol, ('127.0.0.1', 1234))
func = sayhi(protocol, ('127.0.0.1', 1234))
loop.run_until_complete(func)

print('done')
# try:
#     loop.run_forever()
# except KeyboardInterrupt:
#     pass

transport.close()
loop.close()



import base64
import concurrent.futures
import hashlib
import json
import os
import subprocess
import time
from concurrent.futures import ThreadPoolExecutor

import requests
from fundrive.core import DriveSystem
from fundrive.download import simple_download
from tqdm import tqdm


class BaseDrive:
    def __init__(self, *args, **kwargs):
        self.session = None
        self.login_anonymous()

    def login_anonymous(self):
        self.session = requests.Session()
        r = self.session.post(
            url='https://www.wenshushu.cn/ap/login/anonymous',
            json={
                "dev_info": "{}"
            }
        )
        self.session.headers['X-TOKEN'] = r.json()['data']['token']
        self.session.headers[
            'User-Agent'] = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:82.0) Gecko/20100101 Firefox/82.0"
        self.session.headers['Accept-Language'] = "en-US, en;q=0.9"

    def storage(self):
        r = self.session.post(
            url='https://www.wenshushu.cn/ap/user/storage',
            json={}
        )
        rsp = r.json()
        rest_space = int(rsp['data']['rest_space'])
        send_space = int(rsp['data']['send_space'])
        storage_space = rest_space + send_space
        print('当前已用空间:{}GB,剩余空间:{}GB,总空间:{}GB'.format(
            round(send_space / 1024 ** 3, 2),
            round(rest_space / 1024 ** 3, 2),
            round(storage_space / 1024 ** 3, 2)
        ))

    def userinfo(self):
        self.session.post(
            url='https://www.wenshushu.cn/ap/user/userinfo',
            json={"plat": "pcweb"}
        )


class Uploader:
    def __init__(self, drive: BaseDrive, file_path, chunk_size=2097152, *args, **kwargs):
        self.drive = drive
        self.session = drive.session
        self.chunk_size = chunk_size
        self.file_path = file_path
        self.file_size = os.path.getsize(file_path)
        self.is_part = True if self.file_size > self.chunk_size else False

    def get_epoch_time(self):
        r = self.session.get(
            url='https://www.wenshushu.cn/ag/time',
            headers={
                "Prod": "com.wenshushu.web.pc",
                "Referer": "https://www.wenshushu.cn/"
            }
        )
        rsp = r.json()
        return rsp["data"]["time"]

    def get_up_id(self, bid: str, ufileid: str, tid: str):
        r = self.session.post(
            url="https://www.wenshushu.cn/ap/uploadv2/getupid",
            json={
                "preid": ufileid,
                "boxid": bid,
                "linkid": tid,
                "utype": "sendcopy",
                "originUpid": "",
                "length": self.file_size,
                "count": 1
            }
        )
        return r.json()["data"]["upId"]

    def psurl(self, fname, upId, psize, part_num=None, pbar: tqdm = None):
        payload = {
            "ispart": self.is_part,
            "fname": fname,
            "fsize": psize,
            "upId": upId,
        }
        if self.is_part:
            payload["partnu"] = part_num
        r = self.session.post(
            url="https://www.wenshushu.cn/ap/uploadv2/psurl",
            json=payload
        )
        rsp = r.json()
        url = rsp["data"]["url"]
        if pbar:
            pbar.update(psize)
        return url

    def file_put(self, psurl_args, fn, offset=0, read_size=None):
        with open(fn, "rb") as fio:
            fio.seek(offset)
            requests.put(url=self.psurl(*psurl_args), data=fio.read(read_size))

    def get_process(self, up_id: str):
        while True:
            r = self.session.post(
                url="https://www.wenshushu.cn/ap/ufile/getprocess",
                json={
                    "processId": up_id
                }
            )
            if r.json()["data"]["rst"] == "success":
                return True
            time.sleep(1)

    def complete(self, is_part, fname, upId, tid, boxid, preid):
        self.session.post(
            url="https://www.wenshushu.cn/ap/uploadv2/complete",
            json={
                "ispart": is_part,
                "fname": fname,
                "upId": upId,
                "location": {
                    "boxid": boxid,
                    "preid": preid
                }
            }
        )
        self.copysend(boxid, tid, preid)

    def addsend(self):
        self.drive.userinfo()
        self.drive.storage()
        epochtime = self.get_epoch_time()
        req_data = {
            "sender": "",
            "remark": "",
            "isextension": False,
            "notSaveTo": False,
            "notDownload": False,
            "notPreview": False,
            "downPreCountLimit": 0,
            "trafficStatus": 0,
            "pwd": "",
            "expire": "1",
            "recvs": ["social", "public"],
            "file_size": self.file_size,
            "file_count": 1
        }
        # POST的内容在服务端会以字串形式接受然后直接拼接X-TOKEN，不会先反序列化JSON字串再拼接
        # 加密函数中的JSON序列化与此处的JSON序列化的字串形式两者必须完全一致，否则校验失败
        r = self.session.post(
            url='https://www.wenshushu.cn/ap/task/addsend',
            json=req_data,
            headers={
                "A-code": self.get_cipherheader(epochtime, self.session.headers['X-TOKEN'], req_data),
                "Prod": "com.wenshushu.web.pc",
                "Referer": "https://www.wenshushu.cn/",
                "Origin": "https://www.wenshushu.cn",
                "Req-Time": epochtime,
            }
        )
        rsp = r.json()
        if rsp["code"] == 1021:
            raise Exception(f'操作太快啦！请{rsp["message"]}秒后重试')

        data = rsp["data"]
        assert data, "需要滑动验证码"
        bid, ufileid, tid = data["bid"], data["ufileid"], data["tid"]
        upId = self.get_up_id(bid, ufileid, tid)
        return bid, ufileid, tid, upId

    def copysend(self, boxid, taskid, preid):
        r = self.session.post(
            url='https://www.wenshushu.cn/ap/task/copysend',
            json={
                'bid': boxid,
                'tid': taskid,
                'ufileid': preid
            }
        )
        rsp = r.json()
        print(f"个人管理链接：{rsp['data']['mgr_url']}")
        print(f"公共链接：{rsp['data']['public_url']}")

    def sha1_str(self, s):
        cm = hashlib.sha1(s.encode()).hexdigest()
        return cm

    def calc_file_hash(self, hash_type, block=None):
        read_size = self.chunk_size if self.is_part else None
        if not block:
            with open(self.file_path, 'rb') as f:
                block = f.read(read_size)
        if hash_type == "MD5":
            hash_code = hashlib.md5(block).hexdigest()
        elif hash_type == "SHA1":
            hash_code = hashlib.sha1(block).hexdigest()
        return hash_code

    def get_cipherheader(self, epochtime, token, data):
        try:
            from Cryptodome.Cipher import DES
            from Cryptodome.Util import Padding
            import base58
        except Exception as e:
            print(e)
            subprocess.check_call(["pip", "install", "pycryptodomex"])
            from Cryptodome.Cipher import DES
            from Cryptodome.Util import Padding
            import base58

        # cipherMethod: DES/CBC/PKCS7Padding
        json_dumps = json.dumps(data, ensure_ascii=False)
        md5_hash_code = hashlib.md5((json_dumps + token).encode()).hexdigest()
        base58_hash_code = base58.b58encode(md5_hash_code)
        # 时间戳逆序取5位并作为时间戳字串索引再次取值，最后拼接"000"
        key_iv = ("".join([epochtime[int(i)] for i in epochtime[::-1][:5]]) + "000").encode()
        cipher = DES.new(key_iv, DES.MODE_CBC, key_iv)
        cipherText = cipher.encrypt(
            Padding.pad(base58_hash_code, DES.block_size, style="pkcs7")
        )
        return base64.b64encode(cipherText)

    def read_file(self):
        part_num = 0
        with open(self.file_path, "rb") as f:
            while True:
                block = f.read(self.chunk_size)
                part_num += 1
                if block:
                    yield block, part_num
                else:
                    return

    def fast(self):
        boxid, preid, taskid, upId = self.addsend()
        cm1, cs1 = self.calc_file_hash("MD5"), self.calc_file_hash("SHA1")
        cm = self.sha1_str(cm1)
        name = os.path.basename(self.file_path)

        payload = {
            "hash": {
                "cm1": cm1,  # MD5
                "cs1": cs1,  # SHA1
            },
            "uf": {
                "name": name,
                "boxid": boxid,
                "preid": preid
            },
            "upId": upId
        }

        if not self.is_part:
            payload['hash']['cm'] = cm  # 把MD5用SHA1加密
        for _ in range(2):
            r = self.session.post(
                url='https://www.wenshushu.cn/ap/uploadv2/fast',
                json=payload
            )
            rsp = r.json()
            can_fast = rsp["data"]["status"]
            ufile = rsp['data']['ufile']
            if can_fast and not ufile:
                hash_codes = ''
                for block, _ in self.read_file():
                    hash_codes += self.calc_file_hash("MD5", block)
                payload['hash']['cm'] = self.sha1_str(hash_codes)
            elif can_fast and ufile:
                print(f'文件{name}可以被秒传！')
                self.get_process(upId)
                self.copysend(boxid, taskid, preid)

        return name, taskid, boxid, preid, upId

    def upload(self, max_workers=os.cpu_count()):
        fname, tid, boxid, preid, upId = self.fast()
        pbar = tqdm(desc=f"{max_workers}%{fname}", total=self.file_size, unit="B", unit_scale=True, unit_divisor=1024)
        if self.is_part:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                future_list = []
                for i in range((self.file_size + self.chunk_size - 1) // self.chunk_size):
                    ul_size = self.chunk_size if self.chunk_size * (i + 1) <= self.file_size \
                        else self.file_size % self.chunk_size
                    future_list.append(executor.submit(
                        self.file_put, [fname, upId, ul_size, i + 1, pbar],
                        self.file_path, self.chunk_size * i, ul_size))
                concurrent.futures.as_completed(future_list)
        else:
            self.file_put([fname, upId, self.file_size, pbar], self.file_path, 0, self.file_size)

        self.complete(self.is_part, fname, upId, tid, boxid, preid)
        self.get_process(upId)


class Downloader:
    def __init__(self, drive: BaseDrive, share_url, cache_dir='./', *args, **kwargs):
        self.drive = drive
        self.session = drive.session
        self.share_url = share_url
        self.cache_dir = cache_dir
        super().__init__(*args, **kwargs)

    def get_tid(self, token):
        r = self.session.post(
            url='https://www.wenshushu.cn/ap/task/token',
            json={
                'token': token
            }
        )
        return r.json()['data']['tid']

    def mgrtask(self, tid):
        r = self.session.post(
            url='https://www.wenshushu.cn/ap/task/mgrtask',
            json={
                'tid': tid,
                'password': ''
            }
        )
        rsp = r.json()
        expire = rsp['data']['expire']
        days, remainder = divmod(int(float(expire)), 3600 * 24)
        hours, remainder = divmod(remainder, 3600)
        minutes, seconds = divmod(remainder, 60)
        print(f'文件过期时间:{days}天{hours}时{minutes}分{seconds}秒')

        file_size = rsp['data']['file_size']
        print(f'文件大小:{round(int(file_size) / 1024 ** 2, 2)}MB')
        return rsp['data']['boxid'], rsp['data']['ufileid']  # pid

    def sign_url(self, fid):
        r = self.session.post(
            url='https://www.wenshushu.cn/ap/dl/sign',
            json={
                'consumeCode': 0,
                'type': 1,
                'ufileid': fid
            }
        )
        if r.json()['data']['url'] == "" and r.json()['data']['ttNeed'] != 0:
            raise Exception("对方的分享流量不足")
        return r.json()['data']['url']

    def download(self):
        url = self.share_url
        if len(url.split('/')[-1]) == 16:
            token = url.split('/')[-1]
            tid = self.get_tid(token)
        elif len(url.split('/')[-1]) == 11:
            tid = url.split('/')[-1]
        else:
            raise Exception('链接错误')
        bid, pid = self.mgrtask(tid)
        r = self.session.post(
            url='https://www.wenshushu.cn/ap/ufile/list',
            json={
                "start": 0,
                "sort": {
                    "name": "asc"
                },
                "bid": bid,
                "pid": pid,
                "type": 1,
                "options": {
                    "uploader": "true"
                },
                "size": 50
            }
        )
        rsp = r.json()
        for file in rsp['data']['fileList']:
            filename = file['fname']
            url = self.sign_url(file['fid'])
            simple_download(url, filepath=f'{self.cache_dir}/{filename}')


class WSSDrive(DriveSystem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.drive = BaseDrive()

    def upload_dir(self, dir_path="./cache", overwrite=False, *args, **kwargs) -> bool:
        pass

    def download_dir(self, dir_path="./cache", overwrite=False, *args, **kwargs) -> bool:
        pass

    def upload_file(self, file_path="./cache", overwrite=False, *args, **kwargs) -> bool:
        uploader = Uploader(drive=self.drive, file_path=file_path)
        try:
            uploader.upload()
        except Exception as e:
            print(e)

    def download_file(self, share_url=None, dir_path="./cache", overwrite=False, *args, **kwargs) -> bool:
        downloader = Downloader(drive=self.drive, share_url=share_url, cache_dir=dir_path)
        try:
            downloader.download()
        except Exception as e:
            print(e)



from typing import Any, List, Dict

import requests
from fundrive.core import DriveSystem
from fundrive.download import simple_download


class TSingHuaDrive(DriveSystem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def __get_url(share_key=None, path=""):
        return f"https://cloud.tsinghua.edu.cn/api/v2.1/share-links/{share_key}/dirents/?path={path}"

    def get_dir_list(self, share_key=None, path="", *args, **kwargs) -> List[Dict[str, Any]]:
        result = []
        with requests.Session() as sess:
            r = sess.get(self.__get_url(share_key, path))
            objects = r.json()["dirent_list"]
            for obj in objects:
                if obj["is_dir"]:
                    result.append(
                        {
                            "name": obj["folder_name"],
                            "time": obj["last_modified"],
                            "size": obj["size"],
                            "path": obj["folder_path"],
                        }
                    )
        return result

    def get_file_list(self, share_key=None, path="", pwd=None, *args, **kwargs) -> List[Dict[str, Any]]:
        result = []
        with requests.Session() as sess:
            r = sess.get(self.__get_url(share_key, path))
            objects = r.json()["dirent_list"]
            for obj in objects:
                if not obj["is_dir"]:
                    result.append(
                        {
                            "name": obj["file_name"],
                            "time": obj["last_modified"],
                            "size": obj["size"],
                            "path": obj["file_path"],
                        }
                    )
        return result

    def download_file(self, dir_path="./cache", overwrite=False, share_key=None, path="", *args, **kwargs) -> bool:
        file_url = f"https://cloud.tsinghua.edu.cn/d/{share_key}/files/?p={path}&dl=1"
        return simple_download(url=file_url, filepath=f"{dir_path}/{path}")

    def download_dir(self, dir_path="./cache", overwrite=False, share_key=None, path="", *args, **kwargs) -> bool:
        file_list = self.get_file_list(share_key=share_key, path=path)
        dir_list = self.get_dir_list(share_key=share_key, path=path)

        for file in file_list:
            self.download_file(dir_path=dir_path, share_key=share_key, path=file["path"], overwrite=overwrite)
        for file in dir_list:
            self.download_dir(dir_path=dir_path, share_key=share_key, path=file["path"], overwrite=overwrite)
        return True


def download(share_key, dir_path=".cache", path="/", is_dir=True, overwrite=False, *args, **kwargs):
    drive = TSingHuaDrive()
    if is_dir:
        drive.download_dir(share_key=share_key, dir_path=dir_path, path=path, overwrite=overwrite, *args, **kwargs)
    else:
        drive.download_file(share_key=share_key, dir_path=dir_path, path=path, overwrite=overwrite, *args, **kwargs)



import logging

import requests
from funsecret import read_secret

logger = logging.getLogger("tikhub")

"""
https://api.tikhub.io/docs
"""


class TikHubAPI:
    def __init__(self, timeout=10):
        self.domain = "https://api.tikhub.io"
        self.timeout = timeout
        self.token: str = ...
        self._headers: dict = ...

    def login(self, username=None, password=None, token=None) -> bool:
        if token is not None:
            self.token = token
            read_secret(cate1="fundrive", cate2="tiktok", cate3="token", value=self.token)
        elif username is not None and password is not None:
            token_expiry_minutes: int = 525600
            url = f"{self.domain}/user/login?token_expiry_minutes={token_expiry_minutes}"

            payload = {"username": username, "password": password}
            response = requests.request("POST", url, data=payload).json()
            self.token = response.get("access_token")
            read_secret(cate1="fundrive", cate2="tiktok", cate3="token", value=self.token)
            logger.info(f"Your token is: {(response.get('access_token', response.get('detail')))}")
        else:
            self.token = read_secret(cate1="fundrive", cate2="tiktok", cate3="token")

        self._headers = {"Cookie": f"Authorization={self.token}", "Authorization": f"Bearer {self.token}"}
        return True

    def _get(self, uri, params=None) -> dict:
        url = f"{self.domain}/{uri}"

        try:
            response = requests.get(url, headers=self._headers, params=params)
            return response.json()
        except Exception as e:
            return {"status": False, "msg": e}

    def user_me(self):
        return self._get(uri="users/me/")

    def promotion_claim(self, promotion_id):
        data = {"promotion_id": promotion_id}
        return self._get(uri="promotion/claim/", params=data)

    def promotion_daily_check_in(self):
        return self._get(uri="promotion/daily_check_in/")


class DouyinTikHubApi(TikHubAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_douyin_video_data(self, url: str = None, video_id: str = None, language="zh") -> dict:
        """
        解析单一抖音视频数据
        """
        data = {"douyin_video_url": url, "video_id": video_id, "language": language}
        return self._get(uri="douyin/video_data/", params=data)

    def get_douyin_user_data(self, url: str = None, user_id: str = None, language="zh") -> dict:
        """
        解析抖音作者数据
        """
        data = {"douyin_user_url": url, "sec_user_id": user_id, "language": language}
        return self._get(uri="douyin/user_data/", params=data)

    def get_douyin_user_profile_videos_data(
        self,
        url: str = None,
        user_id: str = None,
        max_cursor=0,
        min_cursor=0,
        count=20,
        language="zh",
    ) -> dict:
        """
        解析抖音作者数据
        """
        data = {
            "douyin_user_url": url,
            "sec_user_id": user_id,
            "max_cursor": max_cursor,
            "min_cursor": min_cursor,
            "language": language,
            "count": count,
        }
        return self._get(uri="douyin/user_data/", params=data)


class TikTokTikHubApi(TikHubAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_tiktok_video_data(self, url: str = None, video_id: str = None, region="US", language="en") -> dict:
        """
        解析TikTok视频
        """
        data = {"tiktok_video_url": url, "video_id": video_id, "region": region, "language": language}
        return self._get(uri=f"/tiktok/video_data/", params=data)


class XiaohongshuHubApi(TikHubAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_xhs_note_data(self, url: str = None, note_id: str = None) -> dict:
        """
        解析小红书笔记数据。
        :param url: 小红书笔记链接
        :param note_id: 小红书笔记ID
        优先使用note_id，如果没有note_id则使用url, 两个参数二选一。
        :return:
            {
              "status": true,
              "platform": "string",
              "endpoint": "string",
              "total_time": 0,
              "message": "string",
              "next_url": "string",
              "data": {}
            }
        """
        data = {
            "url": url,
            "note_id": note_id,
        }
        return self._get(uri="/xhs/get_note_data/", params=data)

    def get_xhs_user_info(self, user_id) -> dict:
        """
        解析小红书用户数据。
        :param user_id: 用户ID
        :return:
            {
              "status": true,
              "platform": "string",
              "endpoint": "string",
              "total_time": 0,
              "message": "string",
              "next_url": "string",
              "data": {}
            }
        """
        data = {"user_id": user_id}
        return self._get(uri="/xhs/get_user_info/", params=data)

    def get_xhs_user_notes(self, user_id, cursor=None, num=10) -> dict:
        """
        解析小红书用户笔记数据。
        :param user_id: 用户ID
        :param cursor: 游标(从返回数据中获取)
        :param num: 数量
        :return:
            {
              "status": true,
              "platform": "string",
              "endpoint": "string",
              "total_time": 0,
              "message": "string",
              "next_url": "string",
              "data": {}
            }
        """
        data = {"user_id": user_id, "cursor": cursor, "num": num}
        return self._get(uri="/xhs/get_user_notes/", params=data)

    def get_xhs_note_comments(self, note_id, cursor=None, num=10) -> dict:
        """

        :param note_id:
        :param cursor:
        :param num:
        :return:
            {
              "status": true,
              "platform": "string",
              "endpoint": "string",
              "total_time": 0,
              "message": "string",
              "next_url": "string",
              "data": {}
            }
        """
        data = {"user_id": note_id, "cursor": cursor, "num": num}
        return self._get(uri="/xhs/get_note_comments/", params=data)

    def get_xhs_note_sub_comments(self, note_id, root_comment_id, cursor=None, num=10) -> dict:
        """

        :param note_id:
        :param root_comment_id:
        :param cursor:
        :param num:
        :return:
            {
              "status": true,
              "platform": "string",
              "endpoint": "string",
              "total_time": 0,
              "message": "string",
              "next_url": "string",
              "data": {}
            }
        """
        data = {"user_id": note_id, "root_comment_id": root_comment_id, "cursor": cursor, "num": num}
        return self._get(uri="/xhs/get_note_sub_comments/", params=data)


class WeiboHubApi(TikHubAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_weibo_user_info(self, user_id) -> dict:
        """
        解析微博用户信息数据。
        :param user_id: 用户ID/User ID
        :return:
            {
              "status": true,
              "platform": "string",
              "endpoint": "string",
              "total_time": 0,
              "message": "string",
              "next_url": "string",
              "data": {}
            }
        """
        data = {"user_id": user_id}
        return self._get(uri="/weibo/get_user_info/", params=data)

    def get_weibo_user_posts(self, user_id, page=1, feature=None) -> dict:
        """
        解析微博用户发布的微博数据
        :param user_id: 用户ID/User ID
        :param page: 页数/Page
        :param feature: 0-全部 1-原创 2-图片 3-视频 4-音乐 5-商品 6-链接 7-投票
        :return:
            {
              "status": true,
              "platform": "string",
              "endpoint": "string",
              "total_time": 0,
              "message": "string",
              "next_url": "string",
              "data": {}
            }
        """
        data = {"user_id": user_id, "page": page, "feature": feature}
        return self._get(uri="/weibo/get_user_posts/", params=data)


def test1():
    douyin = DouyinTikHubApi()

    douyin.login()
    print(douyin.user_me())
    # print(tikhub.parse_douyin_video("https://www.douyin.com/video/7153585499477757192"))
    print(douyin.get_douyin_video_data(video_id="7153585499477757192"))

    tiktok = TikTokTikHubApi()
    tiktok.login()
    print(tiktok.get_tiktok_video_data("https://www.tiktok.com/@evil0ctal/video/7202594778217844014"))


# test1()



# -*- coding: utf-8 -*-
import os
import os.path

import requests
from .core import Downloader
from .work import WorkerFactory, Worker
from funfile.compress.utils import file_tqdm_bar


class SpiltDownloader(Downloader):
    def __init__(self, blocks_num=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.blocks_num = blocks_num or self.filesize // (100 * 1024 * 1024)

        if not self.check_available():
            print(f"{self.filename} this url not support range requests,set blocks_num=1.")
            self.blocks_num = 1

    def __get_range(self):
        size = int(self.filesize) // self.blocks_num
        range_list = []
        for i in range(self.blocks_num):
            start = i * size

            if i == self.blocks_num - 1:
                end = self.filesize - 1
            else:
                end = start + size
            if i > 0:
                start += 1
            range_list.append((start, end))
        return range_list

    def download(self, worker_num=5, capacity=100, prefix="", overwrite=False, *args, **kwargs):
        if not overwrite and os.path.exists(self.filepath) and self.filesize == os.path.getsize(self.filepath):
            return False

        prefix = prefix if prefix is not None and len(prefix) > 0 else ""

        range_list = self.__get_range()

        cache_dir = f"{self.filepath}.cache"
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        success_files = []
        pbar = file_tqdm_bar(path=self.filepath, total=self.filesize, prefix=f"{prefix}|0/{self.blocks_num}|")

        def update_pbar(total, curser, current):
            pbar.update(current)
            pbar.refresh()

        with WorkerFactory(worker_num=worker_num, capacity=capacity, timeout=1) as pool:
            for index, (start, end) in enumerate(range_list):
                tmp_file = f"{cache_dir}/split-{str(index).zfill(5)}.tmp"
                success_file = f"{tmp_file}.success"

                def finish_callback(worker: Worker, *args, **kwargs):
                    dst = f"{worker.filepath}.success"
                    os.rename(worker.filepath, dst)
                    success_files.append(dst)
                    pbar.set_description(
                        desc=f"{prefix}|{len(success_files)}/{self.blocks_num}|{os.path.basename(self.filepath)}"
                    )

                worker = Worker(
                    url=self.url,
                    range_start=start,
                    range_end=end,
                    filepath=tmp_file,
                    update_callback=update_pbar,
                    finish_callback=finish_callback,
                )
                if os.path.exists(success_file) and os.path.getsize(success_file) == worker.size:
                    finish_callback(worker)
                    pbar.update(worker.size)
                    pbar.refresh()
                    continue
                pool.submit(worker=worker)

        assert len(success_files) == self.blocks_num
        with open(self.filepath, "wb") as fw:
            for file in success_files:
                with open(file, "rb") as fr:
                    fw.write(fr.read())
                    fw.flush()
                os.remove(file)
            os.removedirs(cache_dir)

    def check_available(self) -> bool:
        header = {"Range": f"bytes=0-100"}
        with requests.get(self.url, stream=True, headers=header) as req:
            return req.status_code == 206


def download(url, filepath, overwrite=False, worker_num=5, capacity=100, prefix=""):
    SpiltDownloader(url=url, filepath=filepath, overwrite=overwrite).download(
        worker_num=worker_num, capacity=capacity, prefix=prefix
    )



import subprocess
from typing import List

from aligo import Aligo
from fundrive.core import BaseDrive
from fundrive.core.base import DriveFile
from funsecret import read_secret


class AlipanDrive(BaseDrive):
    def __init__(self, *args, **kwargs):
        super(AlipanDrive, self).__init__(*args, **kwargs)
        self.drive = Aligo()

    def login(
        self, server_url=None, username=None, password=None, *args, **kwargs
    ) -> bool:
        refresh_token = username or read_secret(
            "fundrive", "drives", "alipan", "refresh_token"
        )
        try:
            from aligo import Aligo
        except Exception as e:
            subprocess.check_call(["pip", "install", "fundrive-alipan"])
            from aligo import Aligo
        self.drive = Aligo(refresh_token=refresh_token)
        return True

    def mkdir(self, fid, name, return_if_exist=True, *args, **kwargs) -> str:
        return self.drive.create_folder(parent_file_id=fid, name=name).file_id

    def delete(self, fid, *args, **kwargs) -> bool:
        self.drive.move_file_to_trash(file_id=fid)
        return True

    def exist(self, fid, *args, **kwargs) -> bool:
        return self.drive.get_file(file_id=fid) is not None

    def get_file_list(self, fid="root", *args, **kwargs) -> List[DriveFile]:
        result = []
        for file in self.drive.get_file_list(parent_file_id=fid):
            if file.type == "file":
                result.append(
                    DriveFile(fid=file.file_id, name=file.name, size=file.size)
                )
        return result

    def get_dir_list(self, fid="root", *args, **kwargs) -> List[DriveFile]:
        result = []
        for file in self.drive.get_file_list(parent_file_id=fid):
            if file.type == "folder":
                result.append(
                    DriveFile(fid=file.file_id, name=file.name, size=file.size)
                )
        return result

    def get_file_info(self, fid, *args, **kwargs) -> DriveFile:
        res = self.drive.get_file(file_id=fid)
        return DriveFile(fid=res.file_id, name=res.name, size=res.size)

    def get_dir_info(self, fid, *args, **kwargs) -> DriveFile:
        res = self.drive.get_file(file_id=fid)
        return DriveFile(fid=res.file_id, name=res.name, size=res.size)

    def download_file(self, fid, local_dir, overwrite=False, *args, **kwargs) -> bool:
        self.drive.download_file(file_id=fid, local_folder=local_dir)
        return True

    def upload_file(
        self, local_path, fid, recursion=True, overwrite=False, *args, **kwargs
    ) -> bool:
        self.drive.upload_file(file_path=local_path, parent_file_id=fid)
        return True



from .tsinghua import TSingHuaDrive
from .tsinghua import download as download_tsinghua
from .tianchi import TianChiDrive
from .oss import OSSDrive
from .opendatalab import OpenDataLabDrive
from .lanzou import LanZouDrive, LanZouSnapshot
from .wenshushu import WSSDrive
from .alipan import AlipanDrive
from .os import OSDrive



import asyncio
import ipaddress
import logging
import os
from pathlib import Path

from aiohttp import web

from fundrive.magnet import settings
from fundrive.magnet.core import Magnet2Torrent
from fundrive.magnet.dht.network import Server as DHTServer
from fundrive.magnet.server import routes
from fundrive.magnet.utils import FailedToFetchException


def start_serve(dht_server, torrent_cache_folder=None, debug=False, apikey=None, ip=ipaddress.IPv4Address("0.0.0.0"),
                port=18667, ):
    if not debug:
        stdio_handler = logging.StreamHandler()
        stdio_handler.setLevel(logging.INFO)
        logger = logging.getLogger("aiohttp.access")
        logger.setLevel(logging.INFO)
        logger.addHandler(stdio_handler)

    settings.SERVE_APIKEY = apikey
    settings.DHT_SERVER = dht_server
    settings.TORRENT_CACHE_FOLDER = torrent_cache_folder
    app = web.Application()
    app.add_routes(routes)
    web.run_app(app, host=str(ip), port=port)


def start_fetch(dht_server, torrent_cache_folder=None, dht_state_file=None, magnet=None, ):
    loop = asyncio.get_event_loop()
    m2t = Magnet2Torrent(dht_server=dht_server, torrent_cache_folder=torrent_cache_folder, use_additional_trackers=True)
    try:
        torrent = loop.run_until_complete(m2t.retrieve_torrent(magnet_link=magnet))
    except FailedToFetchException:
        print("Unable to fetch magnet link")
        quit(1)
    else:
        with open(torrent.name, "wb") as f:
            f.write(torrent.data_encode())

        print("Downloaded magnet link into file: {}".format(torrent.name))

    if dht_server and dht_state_file:
        dht_server.save_state(dht_state_file)


def start(torrent_cache_folder=None, debug=False, use_dht=False, dht_state_file=None, dht_port=settings.DHT_PORT,
          dht_ip=ipaddress.IPv4Address("0.0.0.0"), command='serve', apikey=None, ip=ipaddress.IPv4Address("0.0.0.0"),
          port=18667, magnet=None, ):
    if torrent_cache_folder:
        torrent_cache_folder = Path(torrent_cache_folder)

        if not torrent_cache_folder.exists():
            os.makedirs(torrent_cache_folder)

        if not torrent_cache_folder.is_dir():
            print("Path {} exists but is not a folder".format(torrent_cache_folder))
            quit(1)

    if debug:
        logging.basicConfig(level=logging.DEBUG, format="%(asctime)-15s:%(levelname)s:%(name)s:%(lineno)d:%(message)s")

    if use_dht:
        print("Bootstrapping DHT server")
        loop = asyncio.get_event_loop()
        dht_server = DHTServer()

        if dht_state_file and os.path.isfile(dht_state_file):
            dht_server = DHTServer.load_state(dht_state_file)
            loop.run_until_complete(dht_server.listen(dht_port, str(dht_ip)))
        else:
            dht_server = DHTServer()
            loop.run_until_complete(dht_server.listen(dht_port, str(dht_ip)))
            loop.run_until_complete(dht_server.bootstrap(settings.DHT_BOOTSTRAP_NODES))

        if dht_state_file:
            dht_server.save_state_regularly(dht_state_file)
        print("Done bootstrapping DHT server")
    else:
        dht_server = None

    if command == "serve":
        start_serve(dht_server, torrent_cache_folder=torrent_cache_folder, debug=debug, apikey=apikey, ip=ip, port=port)
    elif command == "fetch":
        start_fetch(dht_server, torrent_cache_folder=torrent_cache_folder, dht_state_file=dht_state_file, magnet=magnet)



import os
import subprocess
from typing import List

from fundrive.core import BaseDrive
from fundrive.core import DriveSnapshot
from fundrive.core.base import DriveFile
from funfile.compress import tarfile
from funsecret import read_secret
from tqdm import tqdm


class Task:
    def __init__(self, url, pwd="", path="./download", now_size=0, folder_id=-1):
        self.url = url
        self.pwd = pwd
        self.path = path
        self.now_size = now_size
        self.folder_id = folder_id


class ProgressWrap:
    def __init__(self, callback: tqdm = None):
        self.callback = callback
        self.last_size = 0

    def init(self, file_name, total_size):
        if self.callback is None:
            self.callback = tqdm(
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                desc=file_name,
                total=total_size,
            )

    def update(self, now_size):
        self.callback.update(now_size - self.last_size)
        self.last_size = now_size


class LanZouDrive(BaseDrive):
    def __init__(self, *args, **kwargs):
        super(LanZouDrive, self).__init__(*args, **kwargs)
        self.allow_big_file = False
        self.drive = None

    def ignore_limit(self):
        self.allow_big_file = True

    def instance(self):
        if self.drive is not None:
            return
        try:
            from lanzou.api import LanZouCloud
        except Exception as e:
            subprocess.check_call(["pip", "install", "fundrive-lanzou"])
            from lanzou.api import LanZouCloud
        self.drive = LanZouCloud()

    def login(
        self, cookie=None, ylogin=None, phpdisk_info=None, *args, **kwargs
    ) -> bool:
        self.instance()
        if cookie is None:
            ylogin = ylogin or read_secret("fundrive", "drives", "funlanzou", "ylogin")
            phpdisk_info = phpdisk_info or read_secret(
                "fundrive", "drives", "funlanzou", "phpdisk_info"
            )
            cookie = {
                "ylogin": ylogin,
                "phpdisk_info": phpdisk_info,
            }
        return self.drive.login_by_cookie(cookie) == 0

    def parse_fid_url_pwd(self, path):
        fid = None
        url = None
        pwd = None
        if isinstance(path, int) or "," not in path:
            fid = path
        else:
            url, pwd = path.split(",")
        return fid, url, pwd

    def exist(self, path, *args, **kwargs) -> bool:
        return True

    def mkdir(self, fid, name, url=None, pwd=None, *args, **kwargs) -> bool:
        return self.drive.mkdir(fid, name, *args, **kwargs) == 0

    def delete(self, fid=None, *args, **kwargs) -> bool:
        return self.drive.delete(fid, *args, **kwargs) == 0

    def get_dir_list(self, fid, url=None, pwd=None, *args, **kwargs) -> List[DriveFile]:
        result = []
        for item in self.drive.get_dir_list(folder_id=fid)[0]:
            result.append(DriveFile(fid=item.id, name=item.name, desc=item.desc))
        return result

    def get_file_list(
        self, fid, url=None, pwd=None, *args, **kwargs
    ) -> List[DriveFile]:
        from lanzou.api.utils import convert_file_size_to_int

        result = []
        if fid is not None:
            for item in self.drive.get_file_list(folder_id=fid):
                result.append(
                    DriveFile(
                        fid=item.id,
                        name=item.name,
                        desc=item.desc,
                        time=item.time,
                        has_pwd=item.has_pwd,
                    )
                )
        if url is not None:
            data = self.drive.get_folder_info_by_url(url, pwd)
            for item in data.files:
                result.append(
                    DriveFile(
                        fid=item.id,
                        name=item.name,
                        desc=item.desc,
                        time=item.time,
                        size=convert_file_size_to_int(item.size),
                        url=item.url,
                        pwd=item.pwd,
                        has_pwd=item.has_pwd,
                    )
                )
        return result

    def get_file_info(self, fid, url=None, pwd=None, *args, **kwargs) -> DriveFile:
        from lanzou.api.utils import convert_file_size_to_int

        data = None
        if fid is not None:
            data = self.drive.get_file_info_by_id(fid)
        if data is None and url is not None:
            data = self.drive.get_file_info_by_url(url, pwd)
        if data is not None:
            return DriveFile(
                fid=data.id,
                name=data.name,
                desc=data.desc,
                time=data.time,
                size=convert_file_size_to_int(data.size),
                url=data.url,
                pwd=data.pwd,
                has_pwd=data.has_pwd,
            )

    def get_dir_info(self, fid, url=None, pwd=None, *args, **kwargs) -> DriveFile:
        data = None
        if fid is not None:
            data = self.drive.get_folder_info_by_id(fid)
        if data is None and url is not None:
            data = self.drive.get_folder_info_by_url(url, pwd)
        if data is not None:
            data = data.folder
            return DriveFile(
                fid=data.id,
                name=data.name,
                desc=data.desc,
                time=data.time,
                url=data.url,
            )

    def download_file(self, fid, local_dir, overwrite=False, *args, **kwargs) -> bool:
        file_info = self.get_file_info(fid=fid)
        task = Task(url=file_info["url"], pwd=file_info["pwd"], path=local_dir)
        os.makedirs(local_dir, exist_ok=True)
        wrap = ProgressWrap()
        wrap.init(file_info["name"], file_info["size"])

        def clb():
            wrap.update(task.now_size)

        return (
            self.drive.down_file_by_url(
                share_url=file_info["url"], task=task, callback=clb
            )
            == 0
        )

    def download_dir(
        self, fid, local_dir, recursion=True, overwrite=False, *args, **kwargs
    ) -> bool:
        if not self.exist(fid):
            return False
        os.makedirs(local_dir, exist_ok=True)
        for file in self.get_file_list(fid):
            self.download_file(
                fid=file["fid"],
                local_dir=local_dir,
                overwrite=overwrite,
                *args,
                **kwargs,
            )
        if not recursion:
            return True

        for file in self.get_dir_list(fid):
            _local_path = os.path.join(local_dir, file["name"])
            self.download_dir(
                fid=file["fid"],
                local_dir=local_dir,
                overwrite=overwrite,
                recursion=recursion,
                *args,
                **kwargs,
            )

    def upload_file(
        self,
        local_path,
        fid,
        url=None,
        pwd=None,
        recursion=True,
        overwrite=False,
        *args,
        **kwargs,
    ) -> bool:
        task = Task(url=url, pwd=pwd, path=local_path, folder_id=fid)
        wrap = ProgressWrap()
        wrap.init(os.path.basename(local_path), os.stat(local_path).st_size)

        def clb():
            wrap.update(task.now_size)

        return (
            self.drive.upload_file(
                task=task,
                file_path=local_path,
                folder_id=fid,
                callback=clb,
                allow_big_file=self.allow_big_file,
            )[0]
            == 0
        )


class LanZouSnapshot(DriveSnapshot):
    def __init__(self, fid=None, url=None, pwd="", *args, **kwargs):
        super(LanZouSnapshot, self).__init__(*args, **kwargs)
        self.drive = LanZouDrive()
        self.fid = fid
        self.url = url
        self.pwd = pwd

    def delete_outed_version(self):
        datas = self.drive.get_file_list(path=self.fid)
        datas = sorted(datas, key=lambda x: x["name"], reverse=True)
        if len(datas) > self.version_num:
            for i in range(self.version_num, len(datas)):
                self.drive.delete(datas[i]["fid"], is_file=True)

    def update(self, file_path, *args, **kwargs):
        gz_path = self._tar_path(file_path)
        tarfile.file_entar(file_path, gz_path)
        self.drive.login()
        self.drive.upload_file(gz_path, drive_path=self.fid)
        os.remove(gz_path)
        self.delete_outed_version()

    def download(self, dir_path, *args, **kwargs):
        self.drive.instance()
        datas = self.drive.get_file_list(url=self.url, pwd=self.pwd)
        if len(datas) == 0:
            print("没有快照文件")
            return
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        datas = sorted(datas, key=lambda x: x["name"], reverse=True)
        self.drive.download_file(
            dir_path=dir_path,
            url=datas[0]["url"] + "," + datas[0]["pwd"],
            overwrite=True,
        )
        tar_path = f"{dir_path}/{datas[0]['name']}"
        tarfile.file_detar(tar_path)
        os.remove(tar_path)



import os

import requests
from tqdm import tqdm


def download_by_request(url, file_path, size=None, overwrite=False, prefix="") -> bool:
    prefix = f"{prefix}--" if prefix is not None and len(prefix) > 0 else ""
    if not os.path.exists(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))
    if not overwrite and os.path.exists(file_path) and os.path.getsize(file_path) == size:
        return False
    with requests.Session() as sess:
        resp = sess.get(url, stream=True)
        size = size or int(resp.headers.get("content-length", 0))

        if not overwrite and os.path.exists(file_path) and os.path.getsize(file_path) == size:
            return False

        with open(file_path, "wb") as file:
            with tqdm(
                total=size,
                ncols=120,
                desc=f"{prefix}{os.path.basename(file_path)}",
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as bar:
                for data in resp.iter_content(chunk_size=2048):
                    bar.update(file.write(data))
                return True



import logging
import os
import sqlite3
from datetime import datetime
from typing import List

import pandas as pd
from funbuild.shell import run_shell


class BaseTable:
    """
    表维度的底层数据库的通用实现
    """

    def __init__(self, table_name: str = "default_table", columns: List[str] = None):
        """
        初始化一个通用数据库
        :param table_name: 表名
        :param columns: 字段列表
        """
        self.table_name = table_name
        self.columns = columns
        self.logger = logging.getLogger(table_name)
        self.conn = None
        self.cursor = None

    def execute(self, sql, commit=True, *args, **kwargs):
        """
        sql执行核心
        :param commit: 是否需要commit
        :param sql: 执行sql
        :return: 执行结果
        """
        try:
            rows = self.cursor.execute(sql)
            if commit:
                self.conn.commit()
            return rows
        except Exception as e:
            print(f"{sql}  with error:{e}")
            return

    def close(self):
        """
        关闭数据库连接
        """
        self.cursor.close()
        self.conn.close()

    def select_pd(self, sql="select * from table_name"):
        """
        将表数据转成pandas的DataFrame
        :param sql: sql
        :return: DataFrame
        """
        sql = self.sql_format(sql)
        return pd.read_sql(sql, self.conn)

    def insert(self, properties: dict):
        """
        插入单条记录，当表设置唯一键插入时，如果唯一键已存在，则返回
        :param properties: 记录以字典形式保存，key是字段名，value是字段值
        :return: 插入 成功or失败
        """
        properties = self.encode(properties)
        keys, values = self._properties2kv(properties)

        sql = """insert or ignore into {table_name} ({columns}) values ({value})""".format(
            table_name=self.table_name, columns=", ".join(keys), value=", ".join(values)
        )
        return self.execute(sql)

    def insert_list(self, property_list: List[dict]):
        """
        批量插入
        :param property_list:
        :return:
        """
        values = [tuple([properties.get(key, "") for key in self.columns]) for properties in property_list]
        sql = "insert or ignore into {} values ({})".format(self.table_name, ",".join(["?"] * len(self.columns)))

        self.cursor.executemany(sql, values)
        self.conn.commit()
        return True

    def update(self, properties: dict, condition: dict):
        """
        更新数据
        :param properties: 需要更新的字段
        :param condition:  where条件
        :return: 更新 成功or失败
        """
        properties = self.encode(properties)
        equal = self._condition2equal(properties)
        equal2 = self._condition2equal(condition)
        sql = """update  {} set {} where {}""".format(self.table_name, ", ".join(equal), " and ".join(equal2))
        return self.execute(sql)

    def update_or_insert(self, properties: dict, condition: dict = None):
        """
        更新或者插入，首先尝试更新，更新失败则插入
        :param properties: 需要更新的字段
        :param condition:  where条件
        :return: 更新/插入 成功or失败
        """
        up = self.update(properties, condition)
        if up is None or up.rowcount == 0:
            return self.insert(properties)
        else:
            return up

    def decode(self, properties: dict):
        """
        需要子类实现
        有些数据插入时可能需要编码/加密等特殊操作，同时，取数据后需要有对应的解码/解密操作，默认不编码/加密
        :param properties: 记录数据
        :return: 编码/加密后的数据
        """
        return properties

    def encode(self, properties: dict):
        """
        需要子类实现
        有些数据插入时可能需要编码/加密等特殊操作，同时，取数据后需要有对应的解码/解密操作，默认不解码/解密
        :param properties: 编码/加密后记录数据
        :return: 解码/加密后的数据
        """
        return properties

    def count(self, properties: dict):
        """
        满足条件的数据量
        :param properties: 记录数
        :return: 记录条数
        """
        properties = properties or {}
        values = []
        for key in self.columns:
            value = str(properties.get(key, ""))
            if len(value) > 0 and len(key) > 0:
                values.append("{}='{}'".format(key, value.replace("'", "")))

        sql = """select count(1) from {} where {}""".format(self.table_name, " and ".join(values))

        rows = self.execute(sql)
        for row in rows:
            return row[0]
        return 0

    def select_all(self):
        """
        返回全表数据
        """
        return self.select("select * from table_name")

    def select(self, sql=None, condition: dict = None):
        """
        根据sql或者指定条件选择数据
        :param sql: sql
        :param condition: 条件
        :return: 记录list
        """
        if sql is None:
            equal2 = self._condition2equal(condition)
            sql = """select * from {} where {}""".format(self.table_name, " and ".join(equal2))
        else:
            sql = self.sql_format(sql)

        rows = self.execute(sql)
        return [] if rows is None else [dict(zip(self.columns, row)) for row in rows]

    def _properties2kv(self, properties: dict):
        """
        将输入的记录数据转换成表的字段名和字段值数据
        :param properties: 记录数据
        :return: keys and values
        """
        if self.columns is None:
            raise Exception("origin_keys cannot be None")
        keys = []
        values = []
        for key in self.columns:
            value = str(properties.get(key, "")).replace("'", "")
            if len(key) > 0 and len(value) > 0:
                keys.append(key)
                values.append("'{}'".format(value))
        return keys, values

    def _condition2equal(self, properties: dict):
        """
        将输入的记录数据转换成表的字段名和字段值数据的等式
        :param properties: 记录数据
        :return: 等式
        """
        if isinstance(properties, str):
            return properties
        if self.columns is None:
            raise Exception("origin_keys cannot be None")
        equals = []
        for key in self.columns:
            value = properties.get(key, None)

            if len(key) > 0 and value is not None:
                if isinstance(value, str):
                    value = value.replace("'", "")
                    equals.append("{}='{}'".format(key, value))
                else:
                    equals.append("{}={}".format(key, value))
        return equals

    def sql_format(self, sql):
        """
        对sql进行格式化，如一些特定字符串的替换
        :param sql: sql
        :return: 格式化之后的sql
        """
        sql = sql.replace("table_name", self.table_name)
        return sql

    def delete(self, condition=None):
        """
        删除表
        :param condition:
        :return:
        """
        if condition is None:
            sql = "delete from {}".format(self.table_name)
        elif isinstance(condition, str):
            sql = "delete from {} where {}".format(self.table_name, condition)
        elif isinstance(condition, dict):
            sql = "delete from {} where {}".format(self.table_name, self._condition2equal(condition))
        else:
            sql = None
        if sql is not None:
            self.execute(sql)
            self.logger.info("delete records with sql = {}".format(sql))

    def add_columns(self, col_name, col_format: str = "VARCHAR(10)"):
        sql = f"ALTER TABLE table_name ADD COLUMN {col_name} {col_format} "
        sql = self.sql_format(sql)
        return self.execute(sql=sql)


class SqliteTable(BaseTable):
    def __init__(self, db_path, conn=None, lanzou_fid=None, *args, **kwargs):
        super(SqliteTable, self).__init__(*args, **kwargs)
        self.db_path = db_path
        self.logger.info("db path:{}".format(self.db_path))

        self.lanzou_fid = lanzou_fid
        if not os.path.exists(os.path.dirname(self.db_path)):
            os.makedirs(os.path.dirname(self.db_path))
        if lanzou_fid is not None and not os.path.exists(self.db_path):
            self.logger.info("file not exist,load from lanzou.")
            self.db_load()
        self.conn = conn or sqlite3.connect(self.db_path, check_same_thread=False)
        self.cursor = self.conn.cursor()

    def save_and_truncate(self):
        result = pd.read_sql("select * from {}".format(self.table_name), self.conn)

        count = len(result)
        path = "{}/{}-{}-{}".format(
            os.path.dirname(self.db_path), self.table_name, count, datetime.now().strftime("%Y%m%d#%H:%M:%S")
        )
        result.to_csv(path)
        self.logger.info("save to csv:{}->{}".format(count, path))

        self.execute("delete from {}".format(self.table_name))
        self.logger.info("delete from {}".format(self.table_name))
        self.vacuum()
        return result

    def to_csv(self, condition, path=None, pop=False, *args, **kwargs):
        if condition is None:
            sql = "select * from {}".format(self.table_name)
        else:
            sql = "select * from {} where {}".format(self.table_name, condition)

        path = path or (
            "{}/{}-{}".format(
                os.path.dirname(self.db_path), self.table_name, datetime.now().strftime("%Y%m%d#%H:%M:%S")
            )
        )
        self.logger.info("save to csv -> {}".format(path))

        cmd = 'sqlite3 -header -csv {db_path} "{sql};" > {path}'.format(db_path=self.db_path, path=path, sql=sql)
        run_shell(cmd)
        if pop:
            self.delete(condition)
            self.vacuum()
        return path

    def pop_to_csv(self, condition, path=None):
        return self.to_csv(condition, path=path, pop=True)

    def vacuum(self):
        """
        数据库清理
        """
        self.execute("VACUUM")
        self.logger.info("数据库VACUUM")

    def insert_list(self, property_list: List[dict]):
        """
        批量插入
        :param property_list:
        :return:
        """
        values = [tuple([properties.get(key, "") for key in self.columns]) for properties in property_list]
        sql = "insert or ignore into {} values ({})".format(self.table_name, ",".join(["?"] * len(self.columns)))

        self.cursor.executemany(sql, values)
        self.conn.commit()
        return True

    def db_load(self):
        from fundrive.lanzou import LanZouCloud

        downer = LanZouCloud()
        downer.ignore_limits()
        downer.login_by_cookie()
        files = downer.get_file_list(folder_id=self.lanzou_fid)
        if len(files) > 0:
            files = sorted(files, key=lambda x: x["id"])
            downer.down_file_by_id(
                files[-1]["id"], save_path=os.path.dirname(self.db_path), file_name=os.path.basename(self.db_path)
            )

    def db_save(self):
        from fundrive.lanzou import LanZouCloud

        downer = LanZouCloud()
        downer.ignore_limits()
        downer.login_by_cookie()
        # file_name = f'{datetime.now().strftime("%Y%m%d%H%M%S")}-{os.path.basename(self.db_path)}'
        file_name = f'db-{datetime.now().strftime("%Y%m%d%H%M%S")}-{os.path.basename(self.db_path)}'
        return downer.upload_file(file_path=self.db_path, folder_id=self.lanzou_fid, file_name=file_name)


class MysqlTable(BaseTable):
    def __init__(self, cursor=None, host=None, user=None, password=None, database=None, *args, **kwargs):
        super(MysqlTable, self).__init__(*args, **kwargs)
        import pymysql

        self.conn = cursor or pymysql.connect(host=host, user=user, password=password, database=database)
        self.cursor = self.conn.cursor()



import os
from typing import List


class DriveFile(dict):
    def __init__(self, fid, name, *args, **kwargs):
        kwargs.update({"fid": fid, "name": name})
        super().__init__(*args, **kwargs)
        self.fid = fid
        self.name = name
        self.data = {"fid": fid, "name": name}

    @property
    def size(self):
        return len(self.data)


class BaseDrive:
    def __init__(self, *args, **kwargs):
        pass

    def login(self, *args, **kwargs) -> bool:
        """
        登录
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def mkdir(self, fid, name, return_if_exist=True, *args, **kwargs) -> str:
        """
        创建目录，如果目录存在则返回目录id
        :param return_if_exist:
        :param fid:
        :param name:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def exist(self, fid, *args, **kwargs) -> bool:
        raise NotImplementedError()

    def delete(self, fid, *args, **kwargs) -> bool:
        """
        删除
        :param fid:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def get_file_list(self, fid, *args, **kwargs) -> List[DriveFile]:
        """
        获取文件列表
        :param fid:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def get_dir_list(self, fid, *args, **kwargs) -> List[DriveFile]:
        """
        获取目录列表
        :param fid:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def get_file_info(self, fid, *args, **kwargs) -> DriveFile:
        """
        获取文件信息
        :param fid:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def get_dir_info(self, fid, *args, **kwargs) -> DriveFile:
        """
        获取目录信息
        :param fid:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def download_file(self, fid, local_dir, overwrite=False, *args, **kwargs) -> bool:
        """
        下载文件
        :param fid:
        :param local_dir:
        :param overwrite:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def download_dir(
        self, fid, local_dir, recursion=True, overwrite=False, *args, **kwargs
    ) -> bool:
        """
        下载目录
        :param fid:
        :param local_dir:
         :param recursion:
        :param overwrite:
        :param args:
        :param kwargs:
        :return:
        """
        if not self.exist(fid):
            return False
        if not os.path.exists(local_dir):
            os.makedirs(local_dir, exist_ok=True)
        for file in self.get_file_list(fid):
            _drive_path = file["fid"]
            self.download_file(
                fid=file.fid, local_dir=local_dir, overwrite=overwrite, *args, **kwargs
            )
        if not recursion:
            return True

        for file in self.get_dir_list(fid):
            self.download_dir(
                fid=file.fid,
                local_dir=os.path.join(local_dir, os.path.basename(fid.name)),
                overwrite=overwrite,
                recursion=recursion,
                *args,
                **kwargs
            )

    def upload_file(
        self, local_path, fid, recursion=True, overwrite=False, *args, **kwargs
    ) -> bool:
        """
        上传文件
        :param local_path:
        :param fid:
        :param recursion:
            :param overwrite:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def upload_dir(
        self, local_path, fid, recursion=True, overwrite=False, *args, **kwargs
    ) -> bool:
        """
        上传目录
        :param local_path:
        :param fid:
         :param recursion:
        :param overwrite:
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()



from sqlalchemy import create_engine as create_engine2

engine_map = {}


def create_engine(uri, cache=True, *args, **kwargs):
    global engine_map
    if cache:
        if uri not in engine_map.keys():
            engine_map[uri] = create_engine2(uri)
        return engine_map[uri]
    return create_engine2(uri)


def create_engine_sqlite(db_path):
    return create_engine(f"sqlite:///{db_path}")


def create_engine_mysql(host, user, password, db_name="", port=3306):
    return create_engine(f"mysql+pymysql://{user}:{password}@{host}:{port}/{db_name}?charset=utf8")



import os
import shutil
from typing import Any, List, Dict

from fundrive.core import BaseDrive


class OSDrive(BaseDrive):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def exist(self, path, *args, **kwargs) -> bool:
        return os.path.exists(path)

    def mkdir(self, path, exist_ok=True, *args, **kwargs) -> bool:
        os.makedirs(path, exist_ok=exist_ok)
        return True

    def upload_file(
        self, local_path, drive_path, recursion=True, overwrite=False, *args, **kwargs
    ) -> bool:
        if not os.path.exists(local_path):
            return False
        print(local_path, drive_path)
        shutil.copyfile(local_path, drive_path)
        return True

    def download_file(
        self, local_path, drive_path, overwrite=False, *args, **kwargs
    ) -> bool:
        if not os.path.exists(drive_path):
            return False
        print(local_path, drive_path)
        shutil.copyfile(drive_path, local_path)
        return True

    def get_file_list(self, path, *args, **kwargs) -> List[Dict[str, Any]]:
        result = []
        for file in os.listdir(path):
            file_path = os.path.join(path, file)
            if os.path.isfile(file_path):
                result.append({"path": file_path})
        return result

    def get_dir_list(self, path, *args, **kwargs) -> List[Dict[str, Any]]:
        result = []
        for file in os.listdir(path):
            file_path = os.path.join(path, file)
            if os.path.isdir(file_path):
                result.append({"path": file_path})
        return result



# -*- coding: utf-8 -*-
import os

import requests


class Downloader:
    def __init__(self, url, filepath, overwrite=False, filesize=None):
        self.url = url
        self.filepath = filepath
        self.overwrite = overwrite
        self.filesize = filesize or self.__get_size()
        self.filename = os.path.basename(self.filepath)

    def download(self, *args, **kwargs):
        pass

    def __get_size(self):
        with requests.Session() as sess:
            resp = sess.get(self.url, stream=True)
            return int(resp.headers.get("content-length", 0))



import json
import os.path

from fundrive.core import BaseDrive
from fundrive.drives.alipan.drive import AlipanDrive


class DriveTable:
    def __init__(self, table_fid, drive: BaseDrive, drive_root=""):
        self.table_fid = table_fid
        self.drive_root = drive_root
        self.drive = drive
        self._fid_par_dict = {}
        self._fid_meta = None
        self._fid_meta_partition_json = None

    def cache_partition(self):
        for file in self.drive.get_dir_list(self.table_fid):
            self._fid_par_dict[file["name"]] = file["fid"]

        if "_meta" in self._fid_par_dict.keys():
            self._fid_meta = self._fid_par_dict["_meta"]
        else:
            self._fid_meta = self.drive.mkdir(fid=self.table_fid, name="_meta")

        for file in self.drive.get_file_list(self._fid_meta):
            if file["name"] == "partition.json":
                self._fid_meta_partition_json = file["fid"]

    @property
    def meta_path(self):
        if len(self._fid_par_dict) == 0:
            self.cache_partition()

        return self._fid_meta

    def upload(self, file, partition):
        if partition in self._fid_par_dict.keys():
            fid = self._fid_par_dict[partition]
        else:
            fid = self.drive.mkdir(fid=self.table_fid, name=partition)
            self.cache_partition()
        self.drive.upload_file(fid=fid, local_path=file)

    def partition_meta_path(self):
        return os.path.join(self.meta_path, "partition.json")

    def update_partition_meta(self, refresh=False, *args, **kwargs):
        tmp = "./partition.json"
        if self._fid_meta_partition_json is not None:
            self.drive.download_file(fid=self._fid_meta_partition_json, local_dir="./")
        partition_meta = []
        if os.path.exists(tmp):
            with open(tmp, "r") as f:
                partition_meta = json.load(f)

        partition_meta = dict([file["name"], file] for file in partition_meta)
        for partition_name, partition_fid in self._fid_par_dict.items():
            if partition_name.startswith("_"):
                continue

            for file in self.drive.get_file_list(partition_fid):
                partition_meta[file["name"]] = file

        with open(tmp, "w") as f:
            json.dump(list(partition_meta.values()), f)
        self.drive.upload_file(local_path=tmp, fid=self._fid_meta)
        os.remove(tmp)

    def partition_meta(self):
        tmp = "./partition.json"
        if self._fid_meta_partition_json is not None:
            self.drive.download_file(local_dir="./", fid=self._fid_meta_partition_json)
        if not os.path.exists(tmp):
            return None
        with open(tmp, "r") as f:
            return json.load(f)



# -*- coding: utf-8 -*-
import os

import requests
from .core import Downloader
from tqdm import tqdm


class SimpleDownloader(Downloader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def download(self, prefix="", chunk_size=2048, *args, **kwargs) -> bool:
        prefix = f"{prefix}--" if prefix is not None and len(prefix) > 0 else ""
        if not os.path.exists(os.path.dirname(self.filepath)):
            os.makedirs(os.path.dirname(self.filepath))
        if not self.overwrite and os.path.exists(self.filepath) and os.path.getsize(self.filepath) == self.filesize:
            return False
        with requests.Session() as sess:
            resp = sess.get(self.url, stream=True)

            with open(self.filepath, "wb") as file:
                with tqdm(
                    total=self.filesize,
                    ncols=120,
                    desc=f"{prefix}{os.path.basename(self.filepath)}",
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                ) as bar:
                    for data in resp.iter_content(chunk_size=chunk_size):
                        bar.update(file.write(data))
                    return True


def download(url, filepath, overwrite=False, prefix="", chunk_size=2048):
    SimpleDownloader(url=url, filepath=filepath, overwrite=overwrite).download(prefix=prefix, chunk_size=chunk_size)



from datetime import datetime


class DriveSnapshot:
    def __init__(self, version_num=20, *args, **kwargs):
        self.version_num = version_num

    def _tar_path(self, file_path):
        return f"{file_path}-{datetime.now().strftime('%Y%m%d%H%M%S')}.tar"

    def update(self, file_path, *args, **kwargs):
        pass

    def download(self, dir_path, *args, **kwargs):
        pass



import json
import os.path

import requests
from fundrive.core import DriveSystem
from fundrive.download import simple_download
from funsecret import read_secret


class TianChiDrive(DriveSystem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cookies = {}

        self.headers = {"content-type": "application/json"}

    def login(self, cookies=None, headers=None, *args, **kwargs) -> bool:
        self.cookies.update(
            cookies
            or {
                "tc": read_secret(cate1="fundrive", cate2="tianchi", cate3="cookies", cate4="tc"),
                "_csrf": read_secret(cate1="fundrive", cate2="tianchi", cate3="cookies", cate4="_csrf"),
            }
        )
        self.headers.update(
            headers
            or {"csrf-token": read_secret(cate1="fundrive", cate2="tianchi", cate3="headers", cate4="csrf-token")}
        )
        return True

    def __get_dataset_url(self, fid=None):
        response = requests.get(
            url=f"https://tianchi.aliyun.com/api/dataset/getFileDownloadUrl?fileId={fid}", cookies=self.cookies
        )
        return response.json()["data"]

    def download_file(self, dir_path="./cache", fid=None, filename=None, overwrite=False, *args, **kwargs) -> bool:
        return simple_download(url=self.__get_dataset_url(fid), filepath=os.path.join(dir_path, filename))

    def download_dir(self, dir_path="./cache", data_id=75730, overwrite=False, *args, **kwargs) -> bool:
        data = json.dumps({"dataId": data_id})

        response = requests.post(
            url="https://tianchi.aliyun.com/api/notebook/dataDetail",
            cookies=self.cookies,
            headers=self.headers,
            data=data,
        )
        data = response.json()["data"]["datalabFile"]

        for item in data:
            self.download_file(fid=item["id"], dir_path=dir_path, filename=item["filePath"])
        return True



import logging
import os
import uuid
from concurrent.futures import ThreadPoolExecutor

from notetool.secret import get_md5_str
from requests import Session
from requests.adapters import HTTPAdapter
from tqdm import tqdm

import m3u8

logger = logging.getLogger("name")
headers = {
    'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:72.0) Gecko/20100101 Firefox/72.0',
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
    'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
    'Connection': 'keep-alive',
    'Upgrade-Insecure-Requests': '1',
}


def get_session(pool_connections, pool_maxsize, max_retries) -> Session:
    """构造session"""
    session = Session()
    adapter = HTTPAdapter(pool_connections=pool_connections, pool_maxsize=pool_maxsize, max_retries=max_retries)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def get_ts_filename(url):
    file_name = url.split('/')[-1].split('?')[0] + '-' + url.split('?')[1].split('&')[0].split('=')[1]
    return file_name


a = str(uuid.uuid4()).replace("-", "")


class FileMeta:
    def __init__(self, url=None, file_dir="", file_name=None):
        self.url = url
        self.file_name = file_name or get_md5_str(url)
        self.file_dir = file_dir or './cache/'

        self.exit_flag = 0

        self.ts_path = os.path.join(self.file_dir, get_md5_str(url))
        self.ts_list = []
        self.key = ""
        self.load_progress = None

        self.count_complete = 0
        self.count_total = 0
        self.meta = None

        if file_dir and not os.path.isdir(file_dir):
            os.makedirs(file_dir)

        if self.ts_path and not os.path.isdir(self.ts_path):
            os.makedirs(self.ts_path)

        self.init_data()

    def load_done(self):
        pass

    def init_data(self):
        self.meta = m3u8.load(self.url)
        for seg in self.meta.segments:
            self.ts_list.append(seg.absolute_uri)
        self.count_total = len(self.ts_list)
        self.load_progress = tqdm(total=len(self.ts_list))

    def load_update(self, size=1):
        self.load_progress.update(size)

    def merge_file(self, file_dir=None, overwrite=True):
        """
        将TS文件整合在一起
        """
        self.count_complete = 0
        file_dir = file_dir or self.file_dir
        file_output = os.path.join(file_dir, self.file_name + '.mp4')

        with open(file_output, 'wb') as outfile:
            for ts_file in tqdm(self.ts_list, desc=f'{self.file_name}'):
                file_name = get_ts_filename(ts_file)
                file_input = os.path.join(self.ts_path, file_name)

                if not os.path.exists(file_input):
                    continue

                with open(file_input, 'rb') as infile:
                    outfile.write(infile.read())

                # 删除临时ts文件
                os.remove(file_input)
        os.rmdir(self.ts_path)


class M3u8Downloader:
    def __init__(self, session=None, thread_size=64):
        self.thread_size = thread_size
        self.session = session or get_session(30, 30, 3)
        self.thread_pool = ThreadPoolExecutor(max_workers=thread_size, thread_name_prefix="download_")

    def download_thead(self, file: FileMeta, ts_index):
        url = file.ts_list[ts_index]
        file_name = get_ts_filename(url)
        file_path = os.path.join(file.ts_path, file_name)
        if os.path.exists(file_path):
            file.load_update()
            return

        for i in range(5, -1, -1):
            try:
                r = self.session.get(url, timeout=20, headers=headers)
                if not r.ok:
                    continue

                with open(file_path, 'wb') as f:
                    f.write(r.content)
                    file.load_update()
                    break
            except Exception as e:
                logger.info(e)
                pass
            if i == 0:
                print('[FAIL]%s' % url)

    def download(self, file: FileMeta):
        for i in range(0, len(file.ts_list)):
            self.thread_pool.submit(self.download_thead, file, i)

    def wait_shutdown(self):
        self.thread_pool.shutdown(wait=True)




from .drives import LanZouDrive
from .drives import OSSDrive
from .drives import TianChiDrive
from .drives import TSingHuaDrive
from .drives import OpenDataLabDrive
from .fungit import GithubDrive
from .drives import LanZouSnapshot
from .fungit import GiteeDrive
from .drives import AlipanDrive



from typing import Any, List, Dict


class DriveSystem:
    def __init__(self, *args, **kwargs):
        pass

    def login(self, *args, **kwargs) -> bool:
        raise NotImplementedError()

    def mkdir(self, path, *args, **kwargs) -> bool:
        raise NotImplementedError()

    def delete(self, *args, **kwargs) -> bool:
        raise NotImplementedError()

    def get_file_list(self, *args, **kwargs) -> List[Dict[str, Any]]:
        raise NotImplementedError()

    def get_dir_list(self, *args, **kwargs) -> List[Dict[str, Any]]:
        raise NotImplementedError()

    def get_file_info(self, *args, **kwargs) -> Dict[str, Any]:
        raise NotImplementedError()

    def get_dir_info(self, *args, **kwargs) -> Dict[str, Any]:
        raise NotImplementedError()

    def download_file(self, dir_path="./cache", overwrite=False, *args, **kwargs) -> bool:
        raise NotImplementedError()

    def download_dir(self, dir_path="./cache", overwrite=False, *args, **kwargs) -> bool:
        raise NotImplementedError()

    def upload_file(self, file_path="./cache", overwrite=False, *args, **kwargs) -> bool:
        raise NotImplementedError()

    def upload_dir(self, dir_path="./cache", overwrite=False, *args, **kwargs) -> bool:
        raise NotImplementedError()



# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.

ext_x_targetduration = '#EXT-X-TARGETDURATION'
ext_x_media_sequence = '#EXT-X-MEDIA-SEQUENCE'
ext_x_discontinuity_sequence = '#EXT-X-DISCONTINUITY-SEQUENCE'
ext_x_program_date_time = '#EXT-X-PROGRAM-DATE-TIME'
ext_x_media = '#EXT-X-MEDIA'
ext_x_playlist_type = '#EXT-X-PLAYLIST-TYPE'
ext_x_key = '#EXT-X-KEY'
ext_x_stream_inf = '#EXT-X-STREAM-INF'
ext_x_version = '#EXT-X-VERSION'
ext_x_allow_cache = '#EXT-X-ALLOW-CACHE'
ext_x_endlist = '#EXT-X-ENDLIST'
extinf = '#EXTINF'
ext_i_frames_only = '#EXT-X-I-FRAMES-ONLY'
ext_x_bitrate = "#EXT-X-BITRATE"
ext_x_byterange = '#EXT-X-BYTERANGE'
ext_x_i_frame_stream_inf = '#EXT-X-I-FRAME-STREAM-INF'
ext_x_discontinuity = '#EXT-X-DISCONTINUITY'
ext_x_cue_out = '#EXT-X-CUE-OUT'
ext_x_cue_out_cont = '#EXT-X-CUE-OUT-CONT'
ext_x_cue_in = '#EXT-X-CUE-IN'
ext_x_cue_span = '#EXT-X-CUE-SPAN'
ext_x_scte35 = '#EXT-OATCLS-SCTE35'
ext_is_independent_segments = '#EXT-X-INDEPENDENT-SEGMENTS'
ext_x_map = '#EXT-X-MAP'
ext_x_start = '#EXT-X-START'
ext_x_server_control = '#EXT-X-SERVER-CONTROL'
ext_x_part_inf = '#EXT-X-PART-INF'
ext_x_part = '#EXT-X-PART'
ext_x_rendition_report = '#EXT-X-RENDITION-REPORT'
ext_x_skip = '#EXT-X-SKIP'
ext_x_session_data = '#EXT-X-SESSION-DATA'
ext_x_session_key = '#EXT-X-SESSION-KEY'
ext_x_preload_hint = '#EXT-X-PRELOAD-HINT'
ext_x_daterange = "#EXT-X-DATERANGE"
ext_x_gap = "#EXT-X-GAP"



import argparse
import os.path
import re
import sys
import urllib.parse
import urllib.request
from html.parser import HTMLParser

from fundrive.aria2 import Aria2JsonRpc

# --------------------------------- Globals ---------------------------------- #

ID = 'a2jrg-scrape'


# -------------------------------- HTMLParser -------------------------------- #
class LinkScanner(HTMLParser):
    """
    HTML URI scraper.
    """

    def __init__(self, *args, ignore_links=False, ignore_images=False, **kwargs):
        self.links = set()
        self.ignore_links = ignore_links
        self.ignore_images = ignore_images
        super().__init__(*args, **kwargs)

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        try:
            if tag == 'a':
                if not self.ignore_links:
                    self.links.add(attrs['href'])
            elif tag == 'img':
                if not self.ignore_images:
                    self.links.add(attrs['src'])
        except KeyError:
            pass


# ----------------------------- argument parsing ----------------------------- #
def parse_args(args=None):
    """
    Parse command-line arguments.
    """
    parser = argparse.ArgumentParser(description='Scrape URIs from a web page and download them with Aria2.', )
    parser.add_argument('uri', help='The URI to scrape.',
                        )
    parser.add_argument(
        'out', metavar='<output dir>',
        help='The output directory.',
    )
    parser.add_argument(
        '-r', '--regex',
        help='Regular expresssion for filtering URIs.',
    )
    parser.add_argument(
        '--ignore-links', action='store_true',
        help='Ignore links.',
    )
    parser.add_argument(
        '--ignore-images', action='store_true',
        help='Ignore images.',
    )
    parser.add_argument(
        '--same',
        help='Indicate that all URIs point to the same file.',
    )

    group = parser.add_argument_group('server options')
    Aria2JsonRpc.add_server_arguments(group)

    group = parser.add_argument_group('download options')
    group.add_argument(
        '-o', '--options', nargs=argparse.REMAINDER,
        help='Aria2 options as key-value pairs separated by "=".'
    )

    return parser.parse_args(args)


# ------------------------------ miscellaneous ------------------------------- #

def get_uris(webpage_uri, **kwargs):
    """
    An iterator over the URIs found at the given HTML webpage.
    """
    scanner = LinkScanner(
        strict=False,
        **kwargs
    )
    with urllib.request.urlopen(webpage_uri) as handle:
        html = handle.read().decode()
    scanner.feed(html)
    for uri in scanner.links:
        uri = urllib.parse.urljoin(webpage_uri, uri)
        uri = urllib.parse.urldefrag(uri).url
        if uri != webpage_uri:
            yield uri


def queue_uris(a2jr, uris, args):
    """
    Queue the given URIs for download via the Aria2 RPC server.
    """
    options = dict()
    options['dir'] = os.path.abspath(args.out)
    if args.options:
        for option in args.options:
            key, value = option.split('=', 1)
            options[key] = value
    if args.same:
        a2jr.queue_uris(uris, options)
    else:
        for uri in uris:
            a2jr.queue_uris([uri], options)


# ----------------------------------- main ----------------------------------- #

def main(args=None):
    pargs = parse_args(args)

    webpage_uri = urllib.parse.urldefrag(pargs.uri).url
    print("collecting URIs...", end='')
    uris = set(get_uris(
        webpage_uri,
        ignore_links=pargs.ignore_links,
        ignore_images=pargs.ignore_images,
    ))
    print(" {:d}".format(len(uris)))

    if pargs.regex:
        regex = re.compile(pargs.regex)
        uris = set(u for u in uris if regex.search(u))
        print("filtered to {:d}".format(len(uris)))

    a2jr = Aria2JsonRpc.a2jr_from_args(ID, pargs)
    queue_uris(a2jr, uris, pargs)


if __name__ == '__main__':
    try:
        main()
    except (KeyboardInterrupt, BrokenPipeError):
        pass
    except urllib.error.HTTPError as err:
        sys.exit(str(err))



import hashlib
import logging
import os
from queue import Queue
from threading import Thread, Lock, current_thread
from urllib import parse

import numpy as np
from fundrive.download import split_download
from requests import Session
from tqdm import tqdm

logging.basicConfig(format="%(asctime)s - [line:%(lineno)d] - %(levelname)s: %(message)s")


def log(name=None, level=logging.DEBUG):
    logger = logging.getLogger(name)
    logger.setLevel(level)

    return logger


logger = log(__name__)

APP_ID = 266719
APP_ID_PAN_WEB = 250528
# APP_ID = 265486
PAN_HOST = "pan.baidu.com"
BASE_URL_RST = "http://pan.baidu.com/rest/2.0"
BASE_URL_PAN = "https://pan.baidu.com/rest/2.0/pcs"
BASE_URL_CPS = "https://pcs.baidu.com/rest/2.0/pcs"

BASE_URL_CPS_NEW = "https://c.pcs.baidu.com/rest/2.0/pcs"


def info(msg):
    logger.info(msg)


def split_file(source_file, target_dir, max_line=2000000):
    file_name = os.path.basename(source_file)
    flag = 0  # 计数器
    name = 1  # 文件名

    info("开始。。。。。")

    def get_filename():
        return str(target_dir) + file_name + "-split-" + str(name) + ".csv"

    write_file = open(get_filename(), "w+")

    with open(source_file, "r") as f_source:
        for line in f_source:
            flag += 1

            write_file.write(line)

            if flag == max_line:
                info("done " + str(flag) + "\t" + get_filename())
                name += 1
                flag = 0

                write_file.close()
                write_file = open(get_filename(), "w+")
    write_file.close()
    info("done " + str(flag) + "\t" + get_filename())
    info("完成。。。。。")


def get_file_md5(path):
    m = hashlib.md5()
    with open(path, "rb") as f:
        while True:
            data = f.read(1024 * 4)
            if not data:
                break
            m.update(data)

    return m.hexdigest()


class SuperDownloaderBak1(object):
    """HTTP下载较大文件的工具"""

    def __init__(self, url, session, save_path, thread_num=45, queue_size=10, chunk=102400):
        """
        :param url: 资源链接
        :param session: 构造好上下文的session
        :param save_path: 保存路径
        :param thread_num: 下载线程数
        :param queue_size: 队列的大小，默认10
        :param chunk: 每个线程下载的块大小
        """
        self.url = url
        self.session = session
        self.save_path = save_path
        self.thread_num = thread_num
        self.queue = Queue(queue_size)
        self.file_size = self._content_length()
        self.position = 0  # 当前的字节偏移量
        self.chunk = chunk
        self.mutex = Lock()  # 资源互斥锁
        self.flags = [False] * self.thread_num

    def download(self):
        with open(self.save_path, "wb") as fp:
            threads = []
            for i in range(self.thread_num):
                p = Thread(target=self._produce, name="%d" % i)
                threads.append(p)
                p.start()
            c = Thread(target=self._consume, args=(fp,), name="consumer")
            threads.append(c)
            c.start()
            for t in threads:
                t.join()

    def _produce(self):
        interval = [0, 0]
        while True:
            if self.mutex.acquire():
                if self.position > self.file_size - 1:
                    self.flags[int(current_thread().getName())] = True
                    self.mutex.release()
                    return
                interval = (self.position, self.position + self.chunk)
                self.position += self.chunk + 1
                self.mutex.release()
            resp = self.session.get(self.url, headers={"Range": "bytes=%s-%s" % interval})
            if not self.queue.full():
                self.queue.put((interval, resp.content))

    def _consume(self, fp):
        while True:
            if all(self.flags) and self.queue.empty():
                return
            if not self.queue.empty():
                item = self.queue.get()
                fp.seek(item[0][0])
                fp.write(item[1])

    def _content_length(self):
        """
        发送head请求获取content-length
        """
        resp = self.session.head(self.url)
        length = resp.headers.get("content-length")
        if length:
            return int(length)
        else:
            raise Exception("%s don't support Range" % self.url)


class SuperDownloaderM(object):
    """
    HTTP下载较大文件的工具
    """

    def __init__(self, session, thread_num=10, queue_size=10, chunk=1024 * 1024):
        """
        :param session: 构造好上下文的session
        :param thread_num: 下载线程数
        :param queue_size: 队列的大小，默认10
        :param chunk: 每个线程下载的块大小
        """
        # self.url = url
        self.session = session

        self.thread_num = thread_num
        self.queue = Queue(queue_size)

        self.position = 0  # 当前的字节偏移量
        self.chunk = chunk
        self.mutex = Lock()  # 资源互斥锁
        self.flags = [False] * self.thread_num

    def download_list(self, meta_list=None, overwrite=True):
        """
        批量下载多个文件
        :param meta_list:
        :param overwrite: 是否覆盖
        :return:
        """
        for meta in meta_list:
            self.download(meta, overwrite=overwrite)
        pass

    def download(self, meta, single=True, overwrite=True):
        """
        下载单个文件
        如果单个文件大小小于100MB,启动单线程下载，否则启动多线程下载

        :param single:
        :param overwrite: 是否覆盖
        :param meta:       文件参数，需包含下列参数
               size        文件大小  byte
               url         下载路径
               file_name   保存文件名
               local_path  本地保存路径，包含文件名
        :return:
        """
        local_path = meta["local_path"]
        file_name = os.path.basename(local_path)

        self.init_path(meta["local_path"])
        if os.path.exists(local_path):
            if not overwrite and meta["md5"] == get_file_md5(local_path):
                info(file_name + "the same md5,file has been exist!, pass")
                return True
            else:
                os.remove(local_path)

        if single or meta["size"] < 10 * 1024 * 1024:
            split_download(meta["url"], meta["local_path"], overwrite=overwrite)
        else:
            split_download(meta["url"], meta["local_path"], overwrite=overwrite)

    def download_single(self, meta):
        """
        单线程下载单个文件
        :param meta:
        :return:
        """
        url = meta["url"]
        file_size = meta["size"]
        local_path = meta["local_path"]
        file_name = os.path.basename(local_path)

        if os.path.exists(local_path):
            first_byte = os.path.getsize(local_path)
        else:
            first_byte = 0

        self.init_path(local_path)

        position = first_byte

        mode = "ab" if first_byte > 0 else "wb"
        with open(local_path, mode=mode) as f:
            with tqdm(
                initial=np.round(first_byte / self.chunk, 2),
                total=np.round(file_size / self.chunk, 2),
                unit="MB",
                desc=file_name,
            ) as pbar:
                while position < file_size:
                    interval = (position, position + self.chunk)
                    position += self.chunk + 1

                    time = 3
                    while time > 0:
                        try:
                            resp = self.session.get(url, headers={"Range": "bytes=%s-%s" % interval})
                            f.write(resp.content)
                            pbar.update(1)
                            break
                        except Exception as e:
                            time -= 1
                            logger.warning(e)

        return True

    def download_multi(self, meta):
        """
        多线程下载单个文件
        :param meta:
        :return:
        """
        with open(meta["local_path"], "wb") as fp:
            threads = []
            for i in range(self.thread_num):
                p = Thread(target=self._produce, args=(meta,), name="%d" % i)
                threads.append(p)
                p.start()
            c = Thread(
                target=self._consume,
                args=(
                    fp,
                    meta,
                ),
                name="consumer",
            )
            threads.append(c)
            c.start()
            for t in threads:
                t.join()

    def _produce(self, meta):
        interval = [0, 0]

        while True:
            if self.mutex.acquire():
                if self.position > meta["size"] - 1:
                    self.flags[int(current_thread().getName())] = True
                    self.mutex.release()
                    return
                interval = (self.position, self.position + self.chunk)
                self.position += self.chunk + 1
                self.mutex.release()
            time = 3
            while time > 0:
                try:
                    resp = self.session.get(meta["url"], headers={"Range": "bytes=%s-%s" % interval})
                    if not self.queue.full():
                        self.queue.put((interval, resp.content))
                    break
                except Exception as e:
                    time -= 1
                    logger.warning(e)

    def _consume(self, fp, meta):
        with tqdm(total=np.round(meta["size"] / self.chunk), unit="MB", desc=meta["server_filename"]) as pbar:
            while True:
                if all(self.flags) and self.queue.empty():
                    return
                if not self.queue.empty():
                    item = self.queue.get()
                    fp.seek(item[0][0])
                    fp.write(item[1])
                    pbar.update(1)

    @staticmethod
    def init_path(path):
        local_dir = os.path.dirname(path)
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)


class SecretManage(object):
    def __init__(self, key, value=None, path="default", secret_dir="~/.secret", save=True):
        self.key = key
        secret_dir = secret_dir.replace("~", os.environ["HOME"])
        self.secret_path = "{}/{}/.{}".format(secret_dir, path, key)
        self.value = value

        if save:
            self.write()

        if self.value is None:
            self.read()

    def read(self):
        """
        从文件读取
        """
        if self.value is not None:
            return self.value
        try:
            if os.path.exists(self.secret_path):
                self.value = open(self.secret_path).read()
                info("read from local")
        except Exception as e:
            logger.warning("read error ,init {}".format(e))
        return self.value

    def write(self):
        """
        写入到文件
        """
        if self.value is None:
            return
        try:
            secret_dir = os.path.dirname(self.secret_path)
            if not os.path.exists(secret_dir):
                os.makedirs(secret_dir)
            with open(self.secret_path, "w") as f:
                f.write(self.value)
                info("write to local")
        except Exception as e:
            logger.warning("error {}".format(e))

    def delete_key(self):
        """
        删除
        """
        if os.path.exists(self.secret_path):
            os.remove(self.secret_path)


class BaiDuDrive(object):
    """
    A client for the PCS API.
    官方API https://openapi.baidu.com/wiki/index.php?title=docs/pcs/rest/file_data_apis_list
    """

    def __init__(self, bduss=None, session=None, timeout=None, access_token=None, save=True):
        """
        :param bduss: The value of baidu cookie key `BDUSS`.
        """
        self.secret = SecretManage(key="bduss", value=bduss, path="baidu", save=save)

        self.bduss = self.secret.read()
        self.session = session or Session()

        self.access_token = access_token
        self.timeout = timeout

        self.session.headers.update({"Cookie": "BDUSS=%s" % self.bduss})

    def quota(self):
        """
        空间配额信息:获取当前用户空间配额信息。

        :return quota: uint64   空间配额，单位为字节。
        :return used: uint64   已使用空间大小，单位为字节。
        """
        return self.request("GET", "/quota", params=dict(method="info"))

    def upload(self, local_path, yun_path, overwrite=True):
        """
        上传单个文件:上传单个文件。百度PCS服务目前支持最大2G的单个文件上传。
        如需支持超大文件（>2G）的断点续传，请参考下面的“分片文件上传”方法。

        注意： 兼容原有域名pcs.baidu.com；使用新域名c.pcs.baidu.com，则提供更快、更稳定的上传服务。

        :param local_path:     本地文件路径（含上传的文件名称）。
        :param yun_path:       云端文件路径（含上传的文件名称）。
        :param overwrite:      是否覆盖同名文件，默认覆盖
        """
        yun_path, local_path = self.parse_path(yun_path, local_path, isdir=False)
        if overwrite:
            ondup = "overwrite"
        else:
            ondup = "newcopy"
        params = {"method": "upload", "path": yun_path, "ondup": ondup}
        files = {yun_path: open(local_path, "rb")}

        meta = self.meta(yun_path=yun_path)
        if meta is not None and len(meta) > 0:
            if get_file_md5(local_path) == meta[0]["md5"]:
                info(yun_path + "the same md5,file has been exist!, pass")
                return

        res = self.request("POST", "/file", params=params, files=files, base_url=BASE_URL_CPS_NEW)
        info("from {} upload to {} done".format(local_path, yun_path))
        return res

    def upload_dir(self, local_dir, yun_dir, overwrite=True):
        """
        上传文件夹
        :param local_dir: 本地文件夹
        :param yun_dir:   云端文件夹
        :param overwrite: 是否覆盖
        :return:
        """
        yun_dir, local_dir = self.parse_path(yun_dir, local_dir)
        if not os.path.exists(local_dir):
            return True
        self.mkdir(yun_dir)

        for file in os.listdir(local_dir):
            file_path_local = os.path.join(local_dir, file)
            file_path_yun = os.path.join(yun_dir, file)

            if os.path.isdir(file_path_local):
                self.upload_dir(file_path_local, file_path_yun, overwrite=overwrite)
            else:
                self.upload(file_path_local, file_path_yun, overwrite=overwrite)

        return True

    def upload_tmpfile(self, local_path):
        """分片上传—文件分片及上传 :百度PCS服务支持每次直接上传最大2G的单个文件。
        如需支持上传超大文件（>2G），则可以通过组合调用分片文件上传的upload方法和createsuperfile方法实现：

        首先，将超大文件分割为2G以内的单文件，并调用upload将分片文件依次上传；
             其次，调用createsuperfile，完成分片文件的重组。

        除此之外，如果应用中需要支持断点续传的功能，也可以通过分片上传文件并调用createsuperfile接口的方式实现。

        兼容原有域名pcs.baidu.com；使用新域名c.pcs.baidu.com，则提供更快、更稳定的上传服务。

        :param local_path: 本地文件路径（含上传的文件名称）。

        :return md5: string   文件的md5签名。
        """
        params = {"method": "upload", "type": "tmpfile"}

        files = {local_path: open(local_path, "rb")}

        return self.request("POST", "/file", params=params, files=files, base_url=BASE_URL_CPS_NEW)

    def upload_superfile(self, block_list, yun_path, ondup=True):
        """分片上传—合并分片文件  :与分片文件上传的upload方法配合使用，可实现超大文件（>2G）上传，同时也可用于断点续传的场景。


        :param block_list: 数组，数组的取值为子文件内容的MD5；子文件至少两个，最多1024个。
        :param yun_path:  云端文件路径（含上传的文件名称）。
        :param ondup: 是否覆盖同名文件，默认覆盖

        """
        params = {
            "method": "createsuperfile",
            "path": yun_path,
            "param": {"block_list": block_list},
            "ondup": ondup,
        }

        return self.request("POST", "/file", params=params)

    def upload_bigfile(self, local_path, yun_path, ondup=True):
        # TODO 大文件分割单个小文件
        paths = local_path.split("")

        block_list = []
        for path in paths:
            res = self.upload_tmpfile(path)
            block_list.append(res["md5"])
        return self.upload_superfile(block_list, yun_path, ondup)
        pass

    def download(self, yun_path, local_path=None, overwrite=True):
        """下载文件
        当local_path为None时，文件会保存在工作目录下，文件名默认为网盘的文件名。
        当local_path以/结尾时，文件名默认为网盘的文件名
        :param overwrite:     是否覆盖
        :param yun_path:      网盘下的文件路径
        :param local_path:    保存本地的文件路径
        """
        yun_path, local_path = self.parse_path(yun_path, local_path, isdir=False)

        params = {"method": "download", "app_id": APP_ID, "path": yun_path}
        query_string = parse.urlencode(params)

        url = BASE_URL_CPS + "/file?%s" % query_string

        meta = self.meta(yun_path)
        if len(meta) == 0:
            return
        meta = meta[0]
        meta.update({"url": url, "local_path": local_path})
        # super_downloader = SuperDownloaderM(self.session)
        # super_downloader.download(meta, single=False, overwrite=overwrite)

        split_download(meta["url"], meta["local_path"], overwrite=overwrite)
        return True

    def download_dir(self, yun_dir, local_dir=None, overwrite=True):
        """下载文件
        当local_path为None时，文件会保存在工作目录下，文件名默认为网盘的文件名。
        当local_path以/结尾时，文件名默认为网盘的文件名
        :param overwrite:     是否覆盖
        :param yun_dir:       网盘下的文件路径
        :param local_dir:     保存本地的文件路径
        """
        yun_dir, local_dir = self.parse_path(yun_dir, local_dir)

        files = self.list_deep(yun_dir, max_depth=3)

        paths = []
        start_ = "###"
        for file in files:
            if file["isdir"] == 1:
                continue
            path = start_ + file["path"]

            path = path.replace(start_ + yun_dir, local_dir)

            params = {"method": "download", "app_id": APP_ID, "path": file["path"]}
            url = BASE_URL_CPS + "/file?%s" % parse.urlencode(params)

            file.update({"from": file["path"], "url": url, "local_path": path})
            paths.append(file)

        for path in paths:
            super_downloader = SuperDownloaderM(self.session)
            super_downloader.download(path, overwrite=overwrite)

        return True

    def mkdir(self, yun_path):
        """创建目录
        :param yun_path:
        创建目录:为当前用户创建一个目录。
        """
        params = {"method": "mkdir", "path": yun_path}
        return self.request("POST", "/file", params=params)

    def meta(self, yun_path):
        """
        获取单个文件/目录的元信息:获取单个文件或目录的元信息。

        :param yun_path:需要获取文件属性的目录，以/开头的绝对路径。如：/apps/album/a/b/c
         注意：: 路径长度限制为1000   非必传   文件名或路径名开头结尾不能是“.”或空白字符，空白字符包括: \r, \n, \t, 空格, \0, \x0B


        :return fs_id: uint64        文件或目录在PCS的临时唯一标识ID。
        :return path: string         文件或目录的绝对路径。
        :return ctime: uint          文件或目录的创建时间。
        :return mtime: uint          文件或目录的最后修改时间。
        :return block_list: string   文件所有分片的md5数组JSON字符串。
        :return size: uint64         文件大小（byte）。
        :return isdir: uint          是否是目录的标识符：“0”为文件 : “1”为目录
        :return ifhassubdir: uint    是否含有子目录的标识符：   “1”表示有子目录
        :return:
        """
        params = {"method": "meta", "path": yun_path}
        res = self.request("GET", "/file", params=params)
        return res.get("list", [])

    def meta_list(self, paths):
        """
        批量获取文件/目录的元信息:批量获取文件或目录的元信息。

        :param paths:   JSON字符串。
        {
        "list":[
            {
                "path":"/apps/album/a/b/c"
            },
            {
                "path":"/apps/album/a/b/d"
            }]
        }

        注意：: 路径长度限制为1000   非必传   文件名或路径名开头结尾不能是“.”或空白字符，空白字符包括: \r, \n, \t, 空格, \0, \x0B


        :return fs_id: uint64   文件或目录在PCS的临时唯一标识id。
        :return path: string   文件或目录的绝对路径。
        :return ctime: uint   文件或目录的创建时间。
        :return mtime: uint   文件或目录的修改时间。
        :return size: uint64   文件大小（byte）。
        :return block_list: string   文件所有分片的md5数组json字符串。
        :return isdir: uint   是否是目录的标识符：
        :return “0”为文件 : “1”为目录   uint
        :return 否: 是否含有子目录的标识符：   “1”表示有子目录
        :return:
        """

        path_list = []
        for path in paths:
            path_list.append({"path": path})

        params = {"method": "meta", "param": {"list": path_list}}
        res = self.request("POST", "/file", params=params)
        return res.get("list", [])

    def list_deep(self, yun_dir, by="name", order="asc", limit="0-100", depth=0, max_depth=5):
        paths = []

        depth += 1
        if depth > max_depth:
            return paths

        files = self.list(yun_dir, by=by, order=order, limit=limit)
        for file in files:
            file["depth"] = depth
            paths.append(file)
            if file["isdir"] == 1:
                paths.extend(
                    self.list_deep(file["path"], by=by, order=order, limit=limit, depth=depth, max_depth=max_depth)
                )

        return paths

    def list(self, yun_dir, by="name", order="asc", limit="0-100"):
        """
        获取目录下的文件列表:获取目录下的文件列表。

        :param yun_dir:    需要list的目录，以/开头的绝对路径。
        :param by:         排序字段，缺省根据文件类型排序：
                                time（修改时间）
                                name（文件名）
                                size（大小，注意目录无大小）
        :param order:      asc（升序）desc（降序）
        :param limit:      否 返回条目控制，参数格式为：n1-n2。返回结果集的[n1, n2)之间的条目，缺省返回所有条目；n1从0开始。


        :return fs_id: uint64   文件或目录在PCS的临时唯一标识id。
        :return path: string    文件或目录的绝对路径。
        :return ctime: uint     文件或目录的创建时间。
        :return mtime: uint     文件或目录的最后修改时间。
        :return md5: string     文件的md5值。
        :return size: uint64    文件大小（byte）。
        :return isdir: uint     是否是目录的标识符：
        """
        params = {
            "method": "list",
            "path": yun_dir,
            "by": by,
            "order": order,
            "limit": limit,
        }
        res = self.request("GET", "/file", params=params)

        return res.get("list", [])

    def move(self, path_from, path_to):
        """
        移动单个文件/目录:移动单个文件/目录。

        :param path_from:   源文件地址（包括文件名）。
        :param path_to:     目标文件地址（包括文件名）。


        :return  描述: from   是
        :return 执行move操作成功的源文件地址。: to   是

        """
        params = {"method": "move", "from": path_from, "to": path_to}
        return self.request("POST", "/file", params=params)

    def move_list(self, move_list):
        """
        移动单个文件/目录:移动单个文件/目录。

        :param move_list:   源文件地址和目标文件地址对应的列表。
        [
            {
                "from":"/apps/album/a/b/c",
                "to":"/apps/album/b/b/c"
            },{
                "from":"/apps/album/a/b/d",
                "to":"/apps/album/b/b/d"
            }
        ]


        :return  描述: from   是
        :return 执行move操作成功的源文件地址。: to   是

        """
        params = {
            "method": "move",
            "param": move_list,
        }
        return self.request("POST", "/file", params=params)

    def copy(self, path_from, path_to):
        """
        复制单个文件/目录:移动单个文件/目录。

        :param path_from:   源文件地址（包括文件名）。
        :param path_to:     目标文件地址（包括文件名）。


        :return  描述: from   是
        :return 执行move操作成功的源文件地址。: to   是

        """
        params = {"method": "copy", "from": path_from, "to": path_to}
        return self.request("POST", "/file", params=params)

    def copy_list(self, move_list):
        """
        复制单个文件/目录:移动单个文件/目录。

        :param move_list:   源文件地址和目标文件地址对应的列表。
        [
            {
                "from":"/apps/album/a/b/c",
                "to":"/apps/album/b/b/c"
            },{
                "from":"/apps/album/a/b/d",
                "to":"/apps/album/b/b/d"
            }
        ]


        :return  描述: from   是
        :return 执行move操作成功的源文件地址。: to   是

        """
        params = {
            "method": "move",
            "param": move_list,
        }
        return self.request("POST", "/file", params=params)

    def delete(self, yun_path):
        """
        获取单个文件/目录的元信息:获取单个文件或目录的元信息。

        :param yun_path:需要获取文件属性的目录，以/开头的绝对路径。如：/apps/album/a/b/c

        :return:
        """
        params = {"method": "delete", "path": yun_path}
        return self.request("POST", "/file", params=params)

    def delete_list(self, paths):
        """
        批量获取文件/目录的元信息:批量获取文件或目录的元信息。

        :param paths:   JSON字符串。如：
        [
            "/apps/album/a/b/c",
            "/apps/album/a/b/d"
        ]

        :return:
        """

        path_list = []
        for path in paths:
            path_list.append({"path": path})

        params = {"method": "delete", "param": {"list": path_list}}
        return self.request("POST", "/file", params=params)

    def get(self, uri, params, *args, **kwargs):
        return self.request("GET", uri, params=params, *args, **kwargs)

    def request(self, method, uri, headers=None, params=None, data=None, files=None, stream=None, base_url=None):
        if base_url is None:
            base_url = BASE_URL_CPS
        if params is None:
            params = {}

        params.update(
            {
                "app_id": params.get("app_id", None) or APP_ID,
                # 'access_token': self.access_token
            }
        )
        # url = os.path.join(base_url, uri)
        url = base_url + uri
        resp = self.session.request(
            method, url, headers=headers, params=params, data=data, files=files, timeout=self.timeout, stream=stream
        )
        return resp.json()

    @staticmethod
    def parse_path(yun_path, loc_path=None, isdir=True):
        cwd = os.getcwd()

        def last_dir(path):
            if path is None:
                return "default"
            p1, p2 = os.path.split(path)
            if p2 == "":
                return os.path.split(p1)[-1]
            else:
                return p2

        if isdir:
            if loc_path is None:
                loc_path = "{}/{}".format(cwd, last_dir(yun_path))
            elif loc_path[0] != "/":
                loc_path = "{}/{}".format(cwd, loc_path)

            if not os.path.exists(loc_path):
                os.makedirs(loc_path)
        else:
            yun_dir, yun_filename = os.path.split(yun_path)

            # 空路径，保存到当前文件夹下
            if loc_path is None:
                loc_path = "{}/{}".format(cwd, yun_filename)
            else:
                loc_dir, loc_filename = os.path.split(loc_path)

                if loc_dir is None or loc_dir == "":
                    loc_dir = cwd
                # 相对路径，前面补上全路径
                elif loc_dir[0] != "/":
                    loc_dir = "{}/{}".format(cwd, loc_dir)

                # 没有设置文件名，取默认文件名
                if loc_filename == "":
                    loc_filename = yun_filename

                loc_path = os.path.join(loc_dir, loc_filename)

                if not os.path.exists(loc_dir):
                    os.makedirs(loc_dir)

        return yun_path, loc_path

    def offline_task_list(self):
        params = {"method": "list_task", "app_id": APP_ID_PAN_WEB}
        res = self.request("POST", "/services/cloud_dl", params=params, base_url=BASE_URL_RST)
        return res

    def offline_task_add(self, source_url, save_path="/local"):
        params = {"method": "add_task", "source_url": source_url, "save_path": save_path, "app_id": APP_ID_PAN_WEB}

        res = self.request("POST", "/services/cloud_dl", params=params, base_url=BASE_URL_RST)
        return res

    def offline_task_add_list(self, source_urls, save_path="/local"):
        res = []
        for source_url in source_urls:
            res.append(self.offline_task_add(source_url=source_url, save_path=save_path))

        return res



import logging

import pandas as pd
from sqlalchemy import delete, Engine
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import DeclarativeBase, sessionmaker
from sqlalchemy.orm import Session
from sqlalchemy.sql import Insert


@compiles(Insert, "sqlite")
def sqlite_insert_ignore(insert, compiler, **kw):
    return compiler.visit_insert(insert.prefix_with("OR IGNORE"), **kw)


@compiles(Insert, "mysql")
def mysql_insert_ignore(insert, compiler, **kw):
    return compiler.visit_insert(insert.prefix_with("IGNORE"), **kw)


@compiles(Insert, "postgresql")
def postgresql_insert_ignore(insert, compiler, **kw):
    statement = compiler.visit_insert(insert, **kw)
    returning_position = statement.find("RETURNING")
    if returning_position >= 0:
        return statement[:returning_position] + "ON CONFLICT DO NOTHING " + statement[returning_position:]
    else:
        return statement + " ON CONFLICT DO NOTHING"


class Base(DeclarativeBase):
    pass


class BaseTable:
    def __init__(self, MyTable, engine: Engine, *args, **kwargs):
        self.table = MyTable
        self.engine: Engine = engine
        self.session = sessionmaker(self.engine)
        self.create()

    def create(self):
        self.table.metadata.create_all(self.engine)

    def execute(self, stmt):
        with self.engine.begin() as conn:
            return conn.execute(stmt)

    def select_all(self):
        with self.engine.begin() as conn:
            return pd.read_sql_table(self.table.__tablename__, conn)

    def delete_all(self):
        return self.execute(delete(self.table))

    def insert(self, values):
        with Session(self.engine) as session:
            try:
                session.bulk_insert_mappings(self.table, values)
                session.commit()
            except Exception as ex:
                logging.warning(ex)
                session.rollback()

    def update(self, values):
        with Session(self.engine) as session:
            try:
                session.bulk_update_mappings(self.table, values)
                session.commit()
            except Exception as ex:
                logging.warning(ex)
                session.rollback()

    def upsert(self, values):
        self.insert(values)
        self.update(values)



import os.path
import subprocess
from typing import List

from fundrive.core import BaseDrive
from fundrive.core.base import DriveFile
from funsecret import read_secret


class WebDavDrive(BaseDrive):
    def __init__(self, *args, **kwargs):
        super(WebDavDrive, self).__init__(*args, **kwargs)
        self.drive = None

    def login(
        self, server_url=None, username=None, password=None, *args, **kwargs
    ) -> bool:
        server_url = server_url or read_secret(
            "fundrive", "webdav", "alipan", "server_url"
        )
        username = username or read_secret("fundrive", "webdav", "alipan", "username")
        password = password or read_secret("fundrive", "webdav", "alipan", "password")
        if not server_url or not username or not password:
            raise Exception("server_url, username, password is None")
        try:
            from webdav4.client import Client
        except Exception as e:
            subprocess.check_call(["pip", "install", "fundrive-webdav"])
            from webdav4.client import Client
        self.drive = Client(server_url, auth=(username, password))
        return True

    def mkdir(self, fid, name, return_if_exist=True, *args, **kwargs) -> str:
        path = str(os.path.join(fid, name))
        self.drive.mkdir(path=path)
        return path

    def delete(self, fid, *args, **kwargs) -> bool:
        self.drive.remove(path=fid)
        return True

    def exist(self, fid, *args, **kwargs) -> bool:
        return self.drive.exists(fid)

    def get_file_list(self, fid, *args, **kwargs) -> List[DriveFile]:
        result = []
        for file in self.drive.ls(path=fid):
            if file["type"] == "file":
                result.append(
                    DriveFile(
                        fid=file["name"],
                        name=os.path.basename(file["name"]),
                        size=file["content_length"],
                    )
                )

        return result

    def get_dir_list(self, fid, *args, **kwargs) -> List[DriveFile]:
        result = []
        for file in self.drive.ls(path=fid):
            if file["type"] == "directory":
                result.append(
                    DriveFile(
                        fid=file["name"],
                        name=os.path.basename(file["name"]),
                        size=file["content_length"],
                    )
                )

        return result

    def get_file_info(self, fid, *args, **kwargs) -> DriveFile:
        res = self.drive.info(fid)
        return DriveFile(
            fid=res["name"],
            name=os.path.basename(res["name"]),
            size=res["content_length"],
        )

    def get_dir_info(self, fid, *args, **kwargs) -> DriveFile:
        res = self.drive.info(fid)
        return DriveFile(
            fid=res["name"],
            name=os.path.basename(res["name"]),
            size=res["content_length"],
        )

    def download_file(self, fid, local_dir, overwrite=False, *args, **kwargs) -> bool:
        local_path = os.path.join(local_dir, os.path.basename(fid))
        os.makedirs(local_dir, exist_ok=True)
        if os.path.exists(local_path) and not overwrite:
            return False
        if not self.exist(fid):
            return False
        self.drive.download_file(from_path=fid, to_path=local_path)
        return True

    def upload_file(
        self, local_path, fid, recursion=True, overwrite=False, *args, **kwargs
    ) -> bool:
        if self.exist(fid) and not overwrite:
            return False
        self.drive.upload_file(
            from_path=local_path,
            to_path=os.path.join(fid, os.path.basename(local_path)),
            overwrite=overwrite,
        )
        return True



# -*- coding: utf-8 -*-
import os
import os.path
import time
from queue import Queue
from threading import Thread
from typing import List

import requests


def _update_callback(total, curser, current):
    """
    :param total: 总大小
    :param curser:当前下载的大小
    :param current: 最新一批次的大小
    :return:
    """
    pass


class Worker:
    def __init__(
        self,
        url: str,
        filepath,
        range_start=0,
        range_end=None,
        update_callback=None,
        finish_callback=None,
        chunk_size=1024 * 1024,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.url = url
        self.filepath = filepath

        self.range_start = range_start
        self.range_curser = range_start
        self.range_end = range_end or self._get_size()
        self.size = self.range_end - self.range_start+1
        self.update_callback = update_callback or _update_callback
        self.finish_callback = finish_callback
        self.chunk_size = chunk_size or 100 * 1024

    def _get_size(self):
        with requests.Session() as sess:
            resp = sess.get(self.url, stream=True)
            return int(resp.headers.get("content-length", 0))

    def run(self):
        if not os.path.exists(os.path.dirname(self.filepath)):
            os.makedirs(os.path.dirname(self.filepath))

        header = {"Range": f"bytes={self.range_curser}-{self.range_end}"}
        with requests.get(self.url, stream=True, headers=header) as req:
            if 200 <= req.status_code <= 299:
                with open(self.filepath, "wb") as cache:
                    for chunk in req.iter_content(chunk_size=self.chunk_size):
                        _size = cache.write(chunk)
                        self.range_curser += _size
                        self.update_callback(self.size, self.range_curser, _size)
        if self.finish_callback:
            self.finish_callback(self)

    def __lt__(self, another):
        return self.range_start < another.range_start

    def get_progress(self):
        """progress for each worker"""
        return {"curser": self.range_curser, "start": self.range_start, "end": self.range_end, "total": self.size}


class _WorkerFactory(object):
    def __init__(self, worker_num, capacity, timeout=30):
        self.worker_num = worker_num
        self._task_queue = Queue(maxsize=capacity)
        self.timeout = timeout
        self._threads: List[Thread] = []
        self._close = False

    def submit(self, worker):
        self._task_queue.put(worker)

    def start(self):
        for i in range(self.worker_num):
            thread = Thread(target=self._worker)
            thread.start()
            self._threads.append(thread)

    def _worker(self):
        while True:
            try:
                worker = self._task_queue.get(timeout=self.timeout)
                worker.run()
                self._task_queue.task_done()
            except Exception as e:
                pass
            if self._close:
                break

    def close(self):
        self._close = True

    def wait_for_all_done(self):
        self._task_queue.join()

    def empty(self):
        return self._task_queue.empty()


class WorkerFactory(object):
    def __init__(self, worker_num=5, capacity=50, timeout=30):
        self.worker_num = worker_num
        self.capacity = capacity
        self.timeout = timeout
        self._handle: _WorkerFactory = ...

    def __enter__(self):
        self._handle = _WorkerFactory(worker_num=self.worker_num, capacity=self.capacity, timeout=self.timeout)
        self._handle.start()
        return self._handle

    def __exit__(self, exc_type, exc_val, exc_tb):
        while not self._handle.empty():
            time.sleep(1)
        self._handle.wait_for_all_done()
        self._handle.close()
        return True



import os
from typing import Any, Dict, List

import requests
from fundrive.core import DriveSystem
from fundrive.download import simple_download
from funsecret import read_secret


class OpenDataLabDrive(DriveSystem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.host = "https://openxlab.org.cn"
        self.cookies = {}
        self.headers = {}

    def login(
        self, ak=None, sk=None, opendatalab_session=None, ssouid=None, *args, **kwargs
    ) -> bool:
        self.cookies.update(
            {
                "opendatalab_session": opendatalab_session
                or read_secret(
                    cate1="fundrive",
                    cate2="opendatalab",
                    cate3="cookies",
                    cate4="opendatalab_session",
                ),
                "ssouid": ssouid
                or read_secret(
                    cate1="fundrive",
                    cate2="opendatalab",
                    cate3="cookies",
                    cate4="ssouid",
                ),
            }
        )
        self.headers.update({"accept": "application/json"})
        return True

    def get_file_info(self, dataset_id, file_path, *args, **kwargs) -> Dict[str, Any]:
        resp = requests.get(
            url=f"{self.host}/datasets/resolve/{dataset_id}/main/{file_path}",
            headers=self.headers,
            cookies=self.cookies,
            allow_redirects=False,
            stream=True,
        )
        result = {
            "url": resp.headers["Location"],
            "dataset_id": dataset_id,
            "path": file_path[1:],
            "size": resp.headers.get("content-length", 0),
        }
        return result

    def get_file_list(
        self, dataset_name, payload=None, *args, **kwargs
    ) -> List[Dict[str, Any]]:
        dataset_name = dataset_name.replace("/", ",")
        data = {"recursive": True}
        if payload:
            data.update(payload)
        resp = requests.get(
            url=f"{self.host}/datasets/api/v2/datasets/{dataset_name}/r/main",
            params=data,
            headers=self.headers,
            cookies=self.cookies,
        )
        result_dict = resp.json()["data"]["list"]
        return result_dict

    def download_file(
        self,
        dir_path="./cache",
        dataset_id=None,
        file_path=None,
        overwrite=False,
        *args,
        **kwargs,
    ) -> bool:
        try:
            file_info = self.get_file_info(dataset_id=dataset_id, file_path=file_path)
            filepath = os.path.join(dir_path, file_info["path"])
            if (
                os.path.exists(filepath)
                and not overwrite
                and file_info["size"] == os.path.getsize(filepath)
            ):
                return False
            return simple_download(
                url=file_info["url"],
                filepath=os.path.join(dir_path, file_info["path"]),
                overwrite=overwrite,
                *args,
                **kwargs,
            )
        except Exception as e:
            return False

    def download_dir(
        self, dir_path="./cache", dataset_name=None, overwrite=False, *args, **kwargs
    ) -> bool:
        if dataset_name is None:
            return False
        file_list = self.get_file_list(dataset_name=dataset_name)
        for i, file in enumerate(file_list):
            filepath = os.path.join(dir_path, file["path"])
            if (
                os.path.exists(filepath)
                and not overwrite
                and file["size"] == os.path.getsize(filepath)
            ):
                return False
            try:
                self.download_file(
                    dir_path=dir_path,
                    dataset_id=file["dataset_id"],
                    file_path=file["path"],
                    overwrite=overwrite,
                    prefix=f"{i}/{len(file_list)}",
                )
            except Exception as e:
                print(e)
        return True



# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.

import datetime
import itertools
import re

import iso8601
from fundrive.m3u8 import protocol

'''
http://tools.ietf.org/html/draft-pantos-http-live-streaming-08#section-3.2
http://stackoverflow.com/questions/2785755/how-to-split-but-ignore-separators-in-quoted-strings-in-python
'''
ATTRIBUTELISTPATTERN = re.compile(r'''((?:[^,"']|"[^"]*"|'[^']*')+)''')
URI_PREFIXES = ('https://', 'http://', 's3://', 's3a://', 's3n://')


def cast_date_time(value):
    return iso8601.parse_date(value)


def format_date_time(value):
    return value.isoformat()


class ParseError(Exception):

    def __init__(self, lineno, line):
        self.lineno = lineno
        self.line = line

    def __str__(self):
        return 'Syntax error in manifest on line %d: %s' % (self.lineno, self.line)


def parse(content, strict=False, custom_tags_parser=None):
    '''
    Given a M3U8 playlist content returns a dictionary with all data found
    '''
    data = {
        'media_sequence': 0,
        'is_variant': False,
        'is_endlist': False,
        'is_i_frames_only': False,
        'is_independent_segments': False,
        'playlist_type': None,
        'playlists': [],
        'segments': [],
        'iframe_playlists': [],
        'media': [],
        'keys': [],
        'rendition_reports': [],
        'skip': {},
        'part_inf': {},
        'session_data': [],
        'session_keys': [],
    }

    state = {
        'expect_segment': False,
        'expect_playlist': False,
        'current_key': None,
        'current_segment_map': None,
    }

    lineno = 0
    for line in string_to_lines(content):
        lineno += 1
        line = line.strip()

        if line.startswith(protocol.ext_x_byterange):
            _parse_byterange(line, state)
            state['expect_segment'] = True

        if line.startswith(protocol.ext_x_bitrate):
            _parse_bitrate(line, state)

        elif line.startswith(protocol.ext_x_targetduration):
            _parse_simple_parameter(line, data, float)

        elif line.startswith(protocol.ext_x_media_sequence):
            _parse_simple_parameter(line, data, int)

        elif line.startswith(protocol.ext_x_discontinuity_sequence):
            _parse_simple_parameter(line, data, int)

        elif line.startswith(protocol.ext_x_program_date_time):
            _, program_date_time = _parse_simple_parameter_raw_value(line, cast_date_time)
            if not data.get('program_date_time'):
                data['program_date_time'] = program_date_time
            state['current_program_date_time'] = program_date_time
            state['program_date_time'] = program_date_time

        elif line.startswith(protocol.ext_x_discontinuity):
            state['discontinuity'] = True

        elif line.startswith(protocol.ext_x_cue_out_cont):
            _parse_cueout_cont(line, state)
            state['cue_out'] = True

        elif line.startswith(protocol.ext_x_cue_out):
            _parse_cueout(line, state, string_to_lines(content)[lineno - 2])
            state['cue_out_start'] = True
            state['cue_out'] = True

        elif line.startswith(protocol.ext_x_cue_in):
            state['cue_in'] = True

        elif line.startswith(protocol.ext_x_cue_span):
            state['cue_out'] = True

        elif line.startswith(protocol.ext_x_version):
            _parse_simple_parameter(line, data, int)

        elif line.startswith(protocol.ext_x_allow_cache):
            _parse_simple_parameter(line, data)

        elif line.startswith(protocol.ext_x_key):
            key = _parse_key(line)
            state['current_key'] = key
            if key not in data['keys']:
                data['keys'].append(key)

        elif line.startswith(protocol.extinf):
            _parse_extinf(line, data, state, lineno, strict)
            state['expect_segment'] = True

        elif line.startswith(protocol.ext_x_stream_inf):
            state['expect_playlist'] = True
            _parse_stream_inf(line, data, state)

        elif line.startswith(protocol.ext_x_i_frame_stream_inf):
            _parse_i_frame_stream_inf(line, data)

        elif line.startswith(protocol.ext_x_media):
            _parse_media(line, data, state)

        elif line.startswith(protocol.ext_x_playlist_type):
            _parse_simple_parameter(line, data)

        elif line.startswith(protocol.ext_i_frames_only):
            data['is_i_frames_only'] = True

        elif line.startswith(protocol.ext_is_independent_segments):
            data['is_independent_segments'] = True

        elif line.startswith(protocol.ext_x_endlist):
            data['is_endlist'] = True

        elif line.startswith(protocol.ext_x_map):
            quoted_parser = remove_quotes_parser('uri')
            segment_map_info = _parse_attribute_list(protocol.ext_x_map, line, quoted_parser)
            state['current_segment_map'] = segment_map_info
            # left for backward compatibility
            data['segment_map'] = segment_map_info

        elif line.startswith(protocol.ext_x_start):
            attribute_parser = {
                "time_offset": lambda x: float(x)
            }
            start_info = _parse_attribute_list(protocol.ext_x_start, line, attribute_parser)
            data['start'] = start_info

        elif line.startswith(protocol.ext_x_server_control):
            _parse_server_control(line, data, state)

        elif line.startswith(protocol.ext_x_part_inf):
            _parse_part_inf(line, data, state)

        elif line.startswith(protocol.ext_x_rendition_report):
            _parse_rendition_report(line, data, state)

        elif line.startswith(protocol.ext_x_part):
            _parse_part(line, data, state)

        elif line.startswith(protocol.ext_x_skip):
            _parse_skip(line, data, state)

        elif line.startswith(protocol.ext_x_session_data):
            _parse_session_data(line, data, state)

        elif line.startswith(protocol.ext_x_session_key):
            _parse_session_key(line, data, state)

        elif line.startswith(protocol.ext_x_preload_hint):
            _parse_preload_hint(line, data, state)

        elif line.startswith(protocol.ext_x_daterange):
            _parse_daterange(line, data, state)

        elif line.startswith(protocol.ext_x_gap):
            state['gap'] = True

        # Comments and whitespace
        elif line.startswith('#'):
            if callable(custom_tags_parser):
                custom_tags_parser(line, data, lineno)

        elif line.strip() == '':
            # blank lines are legal
            pass

        elif state['expect_segment']:
            _parse_ts_chunk(line, data, state)
            state['expect_segment'] = False

        elif state['expect_playlist']:
            _parse_variant_playlist(line, data, state)
            state['expect_playlist'] = False

        elif strict:
            raise ParseError(lineno, line)

    # there could be remaining partial segments
    if 'segment' in state:
        data['segments'].append(state.pop('segment'))

    return data


def _parse_key(line):
    params = ATTRIBUTELISTPATTERN.split(line.replace(protocol.ext_x_key + ':', ''))[1::2]
    key = {}
    for param in params:
        name, value = param.split('=', 1)
        key[normalize_attribute(name)] = remove_quotes(value)
    return key


def _parse_extinf(line, data, state, lineno, strict):
    chunks = line.replace(protocol.extinf + ':', '').split(',', 1)
    if len(chunks) == 2:
        duration, title = chunks
    elif len(chunks) == 1:
        if strict:
            raise ParseError(lineno, line)
        else:
            duration = chunks[0]
            title = ''
    if 'segment' not in state:
        state['segment'] = {}
    state['segment']['duration'] = float(duration)
    state['segment']['title'] = title


def _parse_ts_chunk(line, data, state):
    segment = state.pop('segment')
    if state.get('program_date_time'):
        segment['program_date_time'] = state.pop('program_date_time')
    if state.get('current_program_date_time'):
        segment['current_program_date_time'] = state['current_program_date_time']
        state['current_program_date_time'] += datetime.timedelta(seconds=segment['duration'])
    segment['uri'] = line
    segment['cue_in'] = state.pop('cue_in', False)
    segment['cue_out'] = state.pop('cue_out', False)
    segment['cue_out_start'] = state.pop('cue_out_start', False)
    if state.get('current_cue_out_scte35'):
        segment['scte35'] = state['current_cue_out_scte35']
    if state.get('current_cue_out_duration'):
        segment['scte35_duration'] = state['current_cue_out_duration']
    segment['discontinuity'] = state.pop('discontinuity', False)
    if state.get('current_key'):
        segment['key'] = state['current_key']
    else:
        # For unencrypted segments, the initial key would be None
        if None not in data['keys']:
            data['keys'].append(None)
    if state.get('current_segment_map'):
        segment['init_section'] = state['current_segment_map']
    segment['dateranges'] = state.pop('dateranges', None)
    segment['gap_tag'] = state.pop('gap', None)
    data['segments'].append(segment)


def _parse_attribute_list(prefix, line, atribute_parser):
    params = ATTRIBUTELISTPATTERN.split(line.replace(prefix + ':', ''))[1::2]

    attributes = {}
    for param in params:
        name, value = param.split('=', 1)
        name = normalize_attribute(name)

        if name in atribute_parser:
            value = atribute_parser[name](value)

        attributes[name] = value

    return attributes


def _parse_stream_inf(line, data, state):
    data['is_variant'] = True
    data['media_sequence'] = None
    atribute_parser = remove_quotes_parser('codecs', 'audio', 'video', 'subtitles', 'closed_captions')
    atribute_parser["program_id"] = int
    atribute_parser["bandwidth"] = lambda x: int(float(x))
    atribute_parser["average_bandwidth"] = int
    atribute_parser["frame_rate"] = float
    atribute_parser["video_range"] = str
    atribute_parser["hdcp_level"] = str
    state['stream_info'] = _parse_attribute_list(protocol.ext_x_stream_inf, line, atribute_parser)


def _parse_i_frame_stream_inf(line, data):
    atribute_parser = remove_quotes_parser('codecs', 'uri')
    atribute_parser["program_id"] = int
    atribute_parser["bandwidth"] = int
    atribute_parser["average_bandwidth"] = int
    atribute_parser["video_range"] = str
    atribute_parser["hdcp_level"] = str
    iframe_stream_info = _parse_attribute_list(protocol.ext_x_i_frame_stream_inf, line, atribute_parser)
    iframe_playlist = {'uri': iframe_stream_info.pop('uri'),
                       'iframe_stream_info': iframe_stream_info}

    data['iframe_playlists'].append(iframe_playlist)


def _parse_media(line, data, state):
    quoted = remove_quotes_parser('uri', 'group_id', 'language', 'assoc_language', 'name', 'instream_id',
                                  'characteristics', 'channels')
    media = _parse_attribute_list(protocol.ext_x_media, line, quoted)
    data['media'].append(media)


def _parse_variant_playlist(line, data, state):
    playlist = {'uri': line,
                'stream_info': state.pop('stream_info')}

    data['playlists'].append(playlist)


def _parse_bitrate(line, state):
    if 'segment' not in state:
        state['segment'] = {}
    state['segment']['bitrate'] = line.replace(protocol.ext_x_bitrate + ':', '')


def _parse_byterange(line, state):
    if 'segment' not in state:
        state['segment'] = {}
    state['segment']['byterange'] = line.replace(protocol.ext_x_byterange + ':', '')


def _parse_simple_parameter_raw_value(line, cast_to=str, normalize=False):
    param, value = line.split(':', 1)
    param = normalize_attribute(param.replace('#EXT-X-', ''))
    if normalize:
        value = value.strip().lower()
    return param, cast_to(value)


def _parse_and_set_simple_parameter_raw_value(line, data, cast_to=str, normalize=False):
    param, value = _parse_simple_parameter_raw_value(line, cast_to, normalize)
    data[param] = value
    return data[param]


def _parse_simple_parameter(line, data, cast_to=str):
    return _parse_and_set_simple_parameter_raw_value(line, data, cast_to, True)


def _parse_cueout_cont(line, state):
    elements = line.split(':', 1)
    if len(elements) != 2:
        return
    param, value = elements
    res = re.match('.*Duration=(.*),SCTE35=(.*)$', value)
    if res:
        state['current_cue_out_duration'] = res.group(1)
        state['current_cue_out_scte35'] = res.group(2)


def _cueout_no_duration(line):
    # this needs to be called first since line.split in all other
    # parsers will throw a ValueError if passed just this tag
    if line == protocol.ext_x_cue_out:
        return (None, None)


def _cueout_elemental(line, state, prevline):
    param, value = line.split(':', 1)
    res = re.match('.*EXT-OATCLS-SCTE35:(.*)$', prevline)
    if res:
        return (res.group(1), value)
    else:
        return None


def _cueout_envivio(line, state, prevline):
    param, value = line.split(':', 1)
    res = re.match('.*DURATION=(.*),.*,CUE="(.*)"', value)
    if res:
        return (res.group(2), res.group(1))
    else:
        return None


def _cueout_duration(line):
    # this needs to be called after _cueout_elemental
    # as it would capture those cues incompletely
    # This was added seperately rather than modifying "simple"
    param, value = line.split(':', 1)
    res = re.match(r'DURATION=(.*)', value)
    if res:
        return (None, res.group(1))


def _cueout_simple(line):
    # this needs to be called after _cueout_elemental
    # as it would capture those cues incompletely
    param, value = line.split(':', 1)
    res = re.match(r'^(\d+(?:\.\d)?\d*)$', value)
    if res:
        return (None, res.group(1))


def _parse_cueout(line, state, prevline):
    _cueout_state = (_cueout_no_duration(line)
                     or _cueout_elemental(line, state, prevline)
                     or _cueout_envivio(line, state, prevline)
                     or _cueout_duration(line)
                     or _cueout_simple(line))
    if _cueout_state:
        state['current_cue_out_scte35'] = _cueout_state[0]
        state['current_cue_out_duration'] = _cueout_state[1]


def _parse_server_control(line, data, state):
    attribute_parser = {
        "can_block_reload": str,
        "hold_back": lambda x: float(x),
        "part_hold_back": lambda x: float(x),
        "can_skip_until": lambda x: float(x),
        "can_skip_dateranges": str
    }

    data['server_control'] = _parse_attribute_list(
        protocol.ext_x_server_control, line, attribute_parser
    )


def _parse_part_inf(line, data, state):
    attribute_parser = {
        "part_target": lambda x: float(x)
    }

    data['part_inf'] = _parse_attribute_list(
        protocol.ext_x_part_inf, line, attribute_parser
    )


def _parse_rendition_report(line, data, state):
    attribute_parser = remove_quotes_parser('uri')
    attribute_parser['last_msn'] = int
    attribute_parser['last_part'] = int

    rendition_report = _parse_attribute_list(
        protocol.ext_x_rendition_report, line, attribute_parser
    )

    data['rendition_reports'].append(rendition_report)


def _parse_part(line, data, state):
    attribute_parser = remove_quotes_parser('uri')
    attribute_parser['duration'] = lambda x: float(x)
    attribute_parser['independent'] = str
    attribute_parser['gap'] = str
    attribute_parser['byterange'] = str

    part = _parse_attribute_list(protocol.ext_x_part, line, attribute_parser)

    # this should always be true according to spec
    if state.get('current_program_date_time'):
        part['program_date_time'] = state['current_program_date_time']
        state['current_program_date_time'] += datetime.timedelta(seconds=part['duration'])

    part['dateranges'] = state.pop('dateranges', None)
    part['gap_tag'] = state.pop('gap', None)

    if 'segment' not in state:
        state['segment'] = {}
    segment = state['segment']
    if 'parts' not in segment:
        segment['parts'] = []

    segment['parts'].append(part)


def _parse_skip(line, data, state):
    attribute_parser = remove_quotes_parser('recently_removed_dateranges')
    attribute_parser['skipped_segments'] = int

    data['skip'] = _parse_attribute_list(protocol.ext_x_skip, line, attribute_parser)


def _parse_session_data(line, data, state):
    quoted = remove_quotes_parser('data_id', 'value', 'uri', 'language')
    session_data = _parse_attribute_list(protocol.ext_x_session_data, line, quoted)
    data['session_data'].append(session_data)


def _parse_session_key(line, data, state):
    params = ATTRIBUTELISTPATTERN.split(line.replace(protocol.ext_x_session_key + ':', ''))[1::2]
    key = {}
    for param in params:
        name, value = param.split('=', 1)
        key[normalize_attribute(name)] = remove_quotes(value)
    data['session_keys'].append(key)


def _parse_preload_hint(line, data, state):
    attribute_parser = remove_quotes_parser('uri')
    attribute_parser['type'] = str
    attribute_parser['byterange_start'] = int
    attribute_parser['byterange_length'] = int

    data['preload_hint'] = _parse_attribute_list(
        protocol.ext_x_preload_hint, line, attribute_parser
    )


def _parse_daterange(line, date, state):
    attribute_parser = remove_quotes_parser('id', 'class', 'start_date', 'end_date')
    attribute_parser['duration'] = float
    attribute_parser['planned_duration'] = float
    attribute_parser['end_on_next'] = str
    attribute_parser['scte35_cmd'] = str
    attribute_parser['scte35_out'] = str
    attribute_parser['scte35_in'] = str

    parsed = _parse_attribute_list(
        protocol.ext_x_daterange, line, attribute_parser
    )

    if 'dateranges' not in state:
        state['dateranges'] = []

    state['dateranges'].append(parsed)


def string_to_lines(string):
    return string.strip().splitlines()


def remove_quotes_parser(*attrs):
    return dict(zip(attrs, itertools.repeat(remove_quotes)))


def remove_quotes(string):
    '''
    Remove quotes from string.
    Ex.:
      "foo" -> foo
      'foo' -> foo
      'foo  -> 'foo
    '''
    quotes = ('"', "'")
    if string.startswith(quotes) and string.endswith(quotes):
        return string[1:-1]
    return string


def normalize_attribute(attribute):
    return attribute.replace('-', '_').lower().strip()


def is_url(uri):
    return uri.startswith(URI_PREFIXES)



import os
import subprocess
from typing import Any, List, Dict

from fundrive.core import DriveSystem
from tqdm import tqdm


def public_oss_url(
    bucket_name="nm-algo", endpoint="oss-cn-hangzhou.aliyuncs.com", path=""
):
    return f"https://{bucket_name}.{endpoint}/{path}"


class OSSDrive(DriveSystem):
    def __init__(self, *args, **kwargs):
        super(OSSDrive, self).__init__(*args, **kwargs)
        self.bucket = None

    def login(
        self,
        access_key=None,
        access_secret=None,
        bucket_name=None,
        endpoint=None,
        connect_timeout=3600,
        *args,
        **kwargs,
    ) -> bool:
        try:
            import oss2
        except Exception as e:
            subprocess.check_call(["pip", "install", "fundrive-oss"])
            import oss2

        if access_key is None:
            from funsecret import read_secret

            access_key = read_secret(cate1="fundrive", cate2="oss", cate3="access_key")
            access_secret = read_secret(
                cate1="fundrive", cate2="oss", cate3="access_secret"
            )
            bucket_name = read_secret(
                cate1="fundrive", cate2="oss", cate3="bucket_name"
            )
            endpoint = read_secret(cate1="fundrive", cate2="oss", cate3="endpoint")
        self.bucket = oss2.Bucket(
            oss2.Auth(access_key, access_secret),
            endpoint,
            bucket_name,
            connect_timeout=connect_timeout,
            *args,
            **kwargs,
        )
        return True

    def __get_file_list(self, oss_path) -> List[Dict[str, Any]]:
        result = []
        for file in self.bucket.list_objects(oss_path).object_list:
            result.append(
                {
                    "name": os.path.basename(file.key),
                    "path": file.key,
                    "size": file.size,
                }
            )
        return result

    def get_file_info(self, oss_path, *args, **kwargs) -> Dict[str, Any]:
        result = {}
        files = self.__get_file_list(oss_path=oss_path)
        if len(files) == 1:
            return files[0]
        return result

    def get_dir_info(self, oss_path, *args, **kwargs) -> Dict[str, Any]:
        result = {}
        files = self.__get_file_list(oss_path=oss_path)
        for file in files:
            if file["path"] == oss_path:
                return file
        return result

    def get_file_list(
        self, oss_path, recursion=True, *args, **kwargs
    ) -> List[Dict[str, Any]]:
        result = []
        for file in self.__get_file_list(oss_path):
            if not file["path"].endswith("/"):
                if recursion or len(file["path"].split("/")) == len(
                    oss_path.split("/")
                ):
                    result.append(file)
        return result

    def get_dir_list(
        self, oss_path, recursion=True, *args, **kwargs
    ) -> List[Dict[str, Any]]:
        result = []
        for file in self.__get_file_list(oss_path):
            if file["path"].endswith("/"):
                if (
                    recursion
                    or len(file["path"].split("/")) == len(oss_path.split("/")) + 1
                ):
                    result.append(file)
        return result

    def __download_file(
        self,
        dir_path="./cache",
        oss_path=None,
        size=0,
        overwrite=False,
        *args,
        **kwargs,
    ) -> bool:
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        filename = os.path.basename(oss_path)

        file_path = os.path.join(dir_path, os.path.basename(oss_path))
        if (
            not overwrite
            and os.path.exists(file_path)
            and size == os.path.getsize(file_path)
        ):
            return False

        bar = tqdm(
            total=size,
            ncols=120,
            desc=filename,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        )

        def progress_callback(consumed_bytes, total_bytes):
            bar.update(consumed_bytes - bar.n)

        if not os.path.exists(file_path):
            self.bucket.get_object_to_file(
                oss_path, file_path, progress_callback=progress_callback
            )
        return True

    def download_file(
        self, dir_path="./cache", oss_path=None, overwrite=False, *args, **kwargs
    ) -> bool:
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        file_info = self.get_file_info(oss_path=oss_path)
        return self.__download_file(
            dir_path=dir_path,
            oss_path=oss_path,
            size=file_info["size"],
            overwrite=overwrite,
        )

    def download_dir(
        self, dir_path="./cache", oss_dir=None, overwrite=False, *args, **kwargs
    ) -> bool:
        if oss_dir is None:
            return False
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        for file in self.get_file_list(oss_dir):
            file_path = os.path.join(
                dir_path, os.path.dirname(file["path"].replace(oss_dir, ""))
            )
            self.__download_file(
                dir_path=file_path,
                oss_path=file["path"],
                size=file["size"],
                overwrite=overwrite,
            )
        return True

    def upload_file(
        self, file_path="./cache", oss_dir=None, overwrite=False, *args, **kwargs
    ) -> bool:
        filename = os.path.basename(file_path)
        oss_path = os.path.join(oss_dir, filename)
        size = os.path.getsize(file_path)
        file_info = self.get_file_info(oss_path)
        if not overwrite and "size" in file_info.keys() and size == file_info["size"]:
            return False
        bar = tqdm(
            total=size,
            ncols=120,
            desc=filename,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        )

        def progress_callback(consumed_bytes, total_bytes):
            bar.update(consumed_bytes - bar.n)

        with open(file_path, "rb") as f:
            self.bucket.put_object(oss_path, f, progress_callback=progress_callback)

        return True

    def upload_dir(
        self, dir_path="./cache", oss_dir=None, overwrite=False, *args, **kwargs
    ) -> bool:
        for file in os.listdir(dir_path):
            file_path = os.path.join(dir_path, file)
            if os.path.isfile(file_path):
                self.upload_file(file_path, oss_dir, overwrite=overwrite)
            elif os.path.isdir(file_path):
                self.upload_dir(file_path, os.path.join(oss_dir, file))
        return True
